/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package jlv

import gov.va.med.common.HttpKeys
import gov.va.med.dao.DaoException
import gov.va.med.guibeans.GUIConsult
import gov.va.med.jmeadows.webservice.NoteImage
import gov.va.med.vhahon.common.Utils
import org.apache.commons.beanutils.BeanUtils
import org.apache.commons.lang.StringUtils

import java.nio.charset.Charset

class PatientConsultsController extends PatientNoteDetailBaseController {
    def noteDao
    def auditDao

    static allowedMethods = [list: 'POST', details: 'POST']

    def beforeInterceptor = {

        if (request.patient == null) {
            response.sendError 500
            return false;
        }

        if (actionName == 'details') {
            if (request.recordQuery == null) {
                response.sendError 500
                return false;
            }
        } else dateRangeCheck();
    }

    def list = {

        def guiConsults = noteDao.getPatientConsults(
                session.appUser,
                request.patient,
                Utils.getCal(request[HttpKeys.START_DATE]),
                Utils.getCal(request[HttpKeys.END_DATE]))

        if (guiConsults != null) {
            for (def guiConsult : guiConsults) {
                if (StringUtils.isNotEmpty(guiConsult.getConsultId())) {
                    def queryMap = [(HttpKeys.CONSULT_ID)         : guiConsult.getConsultId(),
                                    (HttpKeys.RECORD_DATE)        : guiConsult.requestDate,
                                    (HttpKeys.RECORD_SITE_AGENCY) : guiConsult.site.agency,
                                    (HttpKeys.RECORD_SITE_MONIKER): guiConsult.site.moniker]

                    guiConsult.setRecordDetailsToken(
                            generateRecordDetailsToken(
                                    params[HttpKeys.PATIENT_TOKEN],
                                    controllerName,
                                    queryMap,
                                    guiConsult.getSite(),
                                    guiConsult.getSourceProtocol(),
                                    guiConsult.isSessionCachedRecord()))

                    if (guiConsult.isSessionCachedRecord()) {
                        //cache consult report in session
                        def consultDetail = BeanUtils.cloneBean(guiConsult)

                        session.appUser.consultMap.put(
                                getCachedRecordKey(guiConsult.getRecordDetailsToken()), consultDetail)

                        //don't send report text with note listings
                        guiConsult.setReport("")
                    }
                }
            }
        }

        sendResponse(guiConsults)
    }

    def details = {

        def guiConsult = null

        if (request.isSessionCachedRecord) {

            //audit details query

            auditDao.auditDetailsQuery("PatConsults", session.appUser, request.patient,
                    request.recordQuery[HttpKeys.RECORD_DATE],
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

            guiConsult = session.appUser.consultMap.get(
                    getCachedRecordKey(params[HttpKeys.RECORD_DETAILS_TOKEN]))
        } else {
            //jMeadows handles audit
            guiConsult = noteDao.getConsultReport(
                    session.appUser,
                    request.patient,
                    request.recordQuery[HttpKeys.CONSULT_ID],
                    request.recordQuery[HttpKeys.RECORD_SITE_CODE])
        }

        if (guiConsult.getReport() != null && guiConsult.getReport().length() == 0) {
            guiConsult.setReport("This report is empty.");
        }
        sendResponse(guiConsult)
    }

    def getNoteDataStream = {

        //retrieve cached note
        GUIConsult guiConsult = session.appUser.consultMap.get(
                getCachedRecordKey(params[HttpKeys.RECORD_DETAILS_TOKEN]))

        def auditQueryName = "GetNoteDataStream"

        if ((guiConsult == null) || !guiConsult.noteAsBinaryStream) {
            response.sendError 500
            return false;
        }

        //stream share/essentris CDA notes
        if (guiConsult.getSite().siteCode.startsWith("2.16.840.1.113883.3.42.10002.100001.8")) {
            auditDao.auditDetailsQuery(auditQueryName, session.appUser, request.patient,
                    request.recordQuery[HttpKeys.RECORD_DATE],
                    request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                    request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

            makeCIS_CDAText(guiConsult.getReport())
            return true;
        }
        //fetch dod note image
        else {
            String complexDataUrl = guiConsult.getComplexDataUrl();

            NoteImage noteImage = null;

            if (complexDataUrl != null) {

                auditDao.auditDetailsQuery(auditQueryName, session.appUser, request.patient,
                        "complexDataUrl:" + complexDataUrl,
                        request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                        request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

                try {
                    noteImage = noteDao.getBHIENoteImage(session.appUser, complexDataUrl);
                }
                catch (DaoException e) {
                    log.error(e);
                }

                response.setHeader("Pragma", "public");
                response.setHeader("Cache-Control", "maxage=1");
                //response.setHeader("Expires", "0");

                if (noteImage != null && noteImage.getNoteBytes() != null) {
                    response.setContentType(noteImage.getContentType());
                    response.setHeader("Content-Disposition", noteImage.getContentDisposition());
                    //response.setContentLength(noteImage.getNoteBytes().length);

//                        //melcolm ** RTF font fix
//                        if(noteImage.getContentType().contains("rtf")){
//                            String rtfString = new String(noteImage.getNoteBytes())
//                            Pattern symbolFont = Pattern.compile("\\{[^\\}]*Symbol;\\}");
//                            rtfString = rtfString.replaceAll(symbolFont, "");
//                            response.getOutputStream() << rtfString.bytes;
//                            response.getOutputStream().flush();
//                            return true;
//                        }
//                        //melcolm ** RTF font fix end

                    //ServletOutputStream out = response.getOutputStream();
//                    if (noteImage.getContentType() != null && noteImage.getContentType().endsWith("/rtf") &&
//                            noteImage.getNoteBytes() != null) {
//                        String rtfStr = new String(noteImage.getNoteBytes());
//                        StringBuilder sb = new StringBuilder();
//
//                        boolean imua = true;
//                        //remove all fonttbl
//                        while (imua) {
//                            int startIndex = rtfStr.indexOf("{\\fonttbl");
//                            if (startIndex < 0) {
//                                sb.append(rtfStr);
//                                //out.print(rtfStr);
//                                imua = false;
//                            } else {
//                                int endIndex = rtfStr.indexOf("}}", startIndex);
//                                sb.append(rtfStr.substring(0, startIndex));
//                                //out.print(rtfStr.substring(0, startIndex));
//                                rtfStr = rtfStr.substring(endIndex + 2, rtfStr.length());
//                            }
//                        }
//                        //response.setContentLength(sb.toString().getBytes(Charset.forName("UTF-8")).length);
//                        //out.print(sb.toString());
//                        response.getOutputStream().write(sb.toString().getBytes(Charset.forName("UTF-8")));
//                    } else {
//                        //response.setContentLength(noteImage.getNoteBytes().length);
                        response.getOutputStream().write(noteImage.getNoteBytes());
//                    }
                    response.getOutputStream().flush();
                    response.getOutputStream().close();
                    return true;
                }
            }

            //if note attachment is null, send an error message

            String errorMsg = "An error occurred while attempting to retrieve note.";

            response.setContentType("text");
            response.setContentLength(errorMsg.length());
            response.getWriter() << errorMsg;
            response.getWriter().flush();

            response.sendError 500
            return false;
        }
    }
}
