/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package gov.va.med.dao;

import gov.va.med.guibeans.*;
import gov.va.med.jmeadows.webservice.Code;
import gov.va.med.jmeadows.webservice.FreeTextReport;
import gov.va.med.jmeadows.webservice.LabOrder;
import gov.va.med.jmeadows.webservice.LabResult;
import gov.va.med.vhahon.common.Utils;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import javax.xml.datatype.DatatypeConstants;
import javax.xml.datatype.XMLGregorianCalendar;
import java.text.SimpleDateFormat;
import java.util.*;

@Component
public class LabDao extends DaoBase {
    private static final Logger logger = Logger.getLogger(LabDao.class.getName());
    private static final SimpleDateFormat sdf = new SimpleDateFormat("MMM dd, yyyy HH:mm");

    public List<GUILabOrder> getPatientLabs(GUIUser user, GUIPatient patient, Calendar startDate, Calendar endDate) {
        try {
            LabOrder[] labs = getServiceHandler().getPatientLabs(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());

            List<GUILabOrder> guiLabs = new ArrayList<GUILabOrder>();
//            SimpleDateFormat sdf = new SimpleDateFormat("MMM dd, yyyy HH:mm");

            for (LabOrder lab : labs) {
                GUILabOrder guiLab = dozerBeanMapper.map(lab, GUILabOrder.class);

                String collectionDate = "";
                String sample = "";
                //String chemId = "";
                String chemORT = "";
                String subscript = "";
                String orderDate = "";
                String orderStatus = "";
                String testName = "";
                String orderComment = "";

                if (lab.getCollectionDate() != null) {
                    synchronized (sdf) {
                        sdf.setTimeZone(lab.getCollectionDate().getTimeZone(DatatypeConstants.FIELD_UNDEFINED));
                        collectionDate = sdf.format(lab.getCollectionDate().toGregorianCalendar().getTime());
                    }
                }

                if (lab.getOrderDate() != null) {
                    synchronized (sdf) {
                        sdf.setTimeZone(lab.getOrderDate().getTimeZone(DatatypeConstants.FIELD_UNDEFINED));
                        orderDate = sdf.format(lab.getOrderDate().toGregorianCalendar().getTime());
                    }
                }

                if (lab.getCollectionSample() != null) {
                    sample = lab.getCollectionSample();
                }

                if (lab.getInternalId() != null) {
                    chemORT = lab.getInternalId();
                }

                if (lab.getStatus() != null) {
                    orderStatus = lab.getStatus();
                }

                if (lab.getSubscript() != null) {
                    subscript = lab.getSubscript();
                }

                if (lab.getRequestingProvider() != null) {
                    guiLab.setRequestingProvider(dozerBeanMapper.map(lab.getRequestingProvider(), GUIProvider.class));
                }

                //CH results contain a panel list
                if (subscript.equalsIgnoreCase("CH") || subscript.contains("CHEM") || subscript.contains("chem")) {
                    guiLab.setFreeTextReport(false);
                } else {
                    guiLab.setFreeTextReport(true);
                }

                guiLab.setCollectionDate(collectionDate);
                guiLab.setSample(sample);
                //guiLab.setChemId(chemId);
                guiLab.setChemORT(chemORT);
                guiLab.setSubscript(subscript);
                guiLab.setLabOrderStatus(orderStatus);
                guiLab.setOrderedDate(orderDate);
                guiLab.setTestName(lab.getTestName());
                guiLab.setComment(lab.getComment());

                if (lab.getLabResults() != null && !lab.getLabResults().isEmpty()) {
                    List<GUILabResult> guiLabResults = new ArrayList<GUILabResult>(lab.getLabResults().size());

                    boolean sensitive = false;
                    for (LabResult labResult : lab.getLabResults()) {
                        GUILabResult guiLabResult = toGUILabResult(labResult);
                        guiLabResult.setOrderDate(orderDate);
                        guiLabResult.setSubscript(guiLab.getSubscript());
                        guiLabResult.setOrderComment(lab.getComment());
                        guiLabResults.add(guiLabResult);
                        if (labResult.isSensitive()) {
                            sensitive = true;
                        }
                    }
                    guiLab.setSensitive(sensitive);
                    guiLab.setLabResults(guiLabResults);
                }

                addSiteDataToGUIBean(guiLab, lab);

                guiLabs.add(guiLab);
            }

            return guiLabs;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public GUILabReport getPatientLabReport(GUIUser user, GUIPatient patient, String collectedDate, String accessionNumber, String recordSiteCode) {
        try {
            Date collectedDateD;
            synchronized (sdf) {
                collectedDateD = sdf.parse(collectedDate);
            }
            GregorianCalendar collectedCalendar = new GregorianCalendar();
            collectedCalendar.setTime(collectedDateD);


            FreeTextReport report = getServiceHandler().getPatientLabReport(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(collectedCalendar)
                            .itemId(accessionNumber)
                            .recordSiteCode(recordSiteCode)
                            .build());

            GUILabReport labReport = new GUILabReport();
            labReport.setReportText(report.getReportText());
            addSiteDataToGUIBean(labReport, report);

            return labReport;
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public List<GUILabResult> getLabOrderResults(GUIUser user, GUIPatient patient, String chemORT, String subscript,
                                                 String recordSiteCode) {
        try {
            LabResult[] labResults = getServiceHandler().getLabOrderResult(

                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .status(subscript)
                            .itemId(chemORT)
                            .recordSiteCode(recordSiteCode)
                            .build());

            List<GUILabResult> guiLabResults = new ArrayList<GUILabResult>();

            for (LabResult labResult : labResults) {
                guiLabResults.add(toGUILabResult(labResult));
            }

            return guiLabResults;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    private GUILabResult toGUILabResult(LabResult labResult) {
        String testId = "";
        String result = "";
        String accession = "";
        String orderId = "";
        String testName = "";
        String printName = "";
        String verifiedBy = "";
        String interpretation = "";
        String referenceRange = "";
        String resultStatus = "";
        String units = "";
        String comment = "";
        String orderComment = "";
        String resultDate = "";
        String collectionDate = "";
        String orderDate = "";
        String specimen = "";
        String subscript = "";
        String patName = "";
        String patToken = "";
        String accessionType = "";
        String performingLabName = "";
        String performingLabAddress = "";
        String requestingLocation = "";

        if (labResult.getTestId() != null) {
            testId = labResult.getTestId();
        }

        if (labResult.getResult() != null) {
            result = labResult.getResult();
        }

        if (labResult.getOrderId() != null) {
            orderId = labResult.getOrderId();
        }

        if (labResult.getTestName() != null) {
            testName = labResult.getTestName();
        }

        if (labResult.getPrintName() != null) {
            printName = labResult.getPrintName();
        }

        if (labResult.getVerifiedBy() != null) {
            verifiedBy = labResult.getVerifiedBy();
        }

        if (labResult.getInterpretation() != null) {
            interpretation = labResult.getInterpretation();
        }

        if (labResult.getReferenceRange() != null) {
            referenceRange = labResult.getReferenceRange();
        }

        if (labResult.getResultStatus() != null) {
            resultStatus = labResult.getResultStatus();
        }

        if (labResult.getUnits() != null) {
            units = labResult.getUnits();
        }

        if (labResult.getComment() != null) {
            comment = labResult.getComment();
        }

        if (labResult.getOrderComment() != null) {
            orderComment = labResult.getOrderComment();
        }

        if(labResult.getPerformingLabName() != null){
            performingLabName = labResult.getPerformingLabName();
        }

        if(labResult.getPerformingLabStreet1() != null){
            performingLabAddress = labResult.getPerformingLabStreet1();
        }

        if(labResult.getRequestingLocation() != null){
            requestingLocation = labResult.getRequestingLocation();
        }

        if (labResult.getResultDate() != null) {
            //resultDate = Utils.formatDisplayDate(labResult.getResultDate());
            synchronized (sdf) {
                sdf.setTimeZone(labResult.getResultDate().getTimeZone(DatatypeConstants.FIELD_UNDEFINED));
                resultDate = sdf.format(labResult.getResultDate().toGregorianCalendar().getTime());
            }
        }

        if (labResult.getCollectedDate() != null) {
            synchronized (sdf) {
                sdf.setTimeZone(labResult.getCollectedDate().getTimeZone(DatatypeConstants.FIELD_UNDEFINED));
                collectionDate = sdf.format(labResult.getCollectedDate().toGregorianCalendar().getTime());
            }
        }

        if (labResult.getOrderDate() != null) {
            synchronized (sdf) {
                sdf.setTimeZone(labResult.getOrderDate().getTimeZone(DatatypeConstants.FIELD_UNDEFINED));
                orderDate = sdf.format(labResult.getOrderDate().toGregorianCalendar().getTime());
            }
        }

        if (labResult.getSpecimen() != null) {
            specimen = labResult.getSpecimen();
        }

        if (labResult.getAccession() != null) {
            accession = labResult.getAccession();

            // the subscript may be embedded in the accession if it is from Vista Data Service (The accession is of the
            // form "XX XX XX^YY" where YY is the subscript.
            int lastCaretIndex = accession.lastIndexOf("^");
            if (lastCaretIndex > 0) {
                subscript = accession.substring(lastCaretIndex + 1);
                accession = accession.substring(0, lastCaretIndex);
            }

            // set the accessionType for VA lab results using the accession number prefix
            if (labResult.getSite() != null &&
                    labResult.getSite().getAgency() != null &&
                    "VA".equalsIgnoreCase(labResult.getSite().getAgency())) {
                int firstSpaceIndex = accession.indexOf(" ", 0);
                if (firstSpaceIndex > 0)
                    accessionType = accession.substring(0, firstSpaceIndex);
            }

        }

        if (labResult.getSite() != null &&
                labResult.getSite().getAgency() != null &&
                "DOD".equalsIgnoreCase(labResult.getSite().getAgency())) {
            if (labResult.getTestType() != null) {
                accessionType = labResult.getTestType();
            }
        }

        if (labResult.getPatientName() != null) {
            patName = labResult.getPatientName();
        }

        GUILabResult guiLabResult = new GUILabResult();

        //graphing only available if cache protocol
        if (appConfig.getProtocolCache().equals(labResult.getSourceProtocol())) {
            try {
                //only allow histogram for numeric results
                Double.parseDouble(result);
                guiLabResult.setHistogramAvailable(true);
            } catch (java.lang.NumberFormatException e) {
                if (logger.isDebugEnabled()) {
                    logger.debug(e.getMessage());
                }
            }
        } else guiLabResult.setHistogramAvailable(false);

        if("VA".equalsIgnoreCase(labResult.getSite().getAgency())){
            if(subscript.equalsIgnoreCase("MI") || accessionType.equalsIgnoreCase("MI") || subscript.equalsIgnoreCase("AP") || accessionType.equalsIgnoreCase("AP")){
                guiLabResult.setFreeTextReport(true);
            } else {
                guiLabResult.setFreeTextReport(false);
            }
        } else {
            //non-CH labs have a free-text report
            if (!subscript.equalsIgnoreCase("CH") && !subscript.contains("CHEM") && !subscript.contains("chem")
                    && !accessionType.equalsIgnoreCase("CH") && !accessionType.contains("CHEM") && !accessionType.contains("chem")) {
                guiLabResult.setFreeTextReport(true);
            }
        }

        if (labResult.getProvider() != null) {
            guiLabResult.setProvider(dozerBeanMapper.map(labResult.getProvider(), GUIProvider.class));
        }

        guiLabResult.setTestId(testId);
        guiLabResult.setSpecimen(specimen);
        guiLabResult.setResult(result);
        guiLabResult.setAccession(accession);
        guiLabResult.setOrderId(orderId);
        guiLabResult.setTestName(testName);
        guiLabResult.setPrintName(printName);
        guiLabResult.setVerifiedBy(verifiedBy);
        guiLabResult.setInterpretation(interpretation);
        guiLabResult.setReferenceRange(referenceRange);
        guiLabResult.setUnits(units);
        guiLabResult.setResultStatus(resultStatus);
        guiLabResult.setComment(comment);
        guiLabResult.setOrderComment(orderComment);
        guiLabResult.setResultDate(resultDate);
        guiLabResult.setCollectionDate(collectionDate);
        guiLabResult.setOrderDate(orderDate);
        guiLabResult.setSpecimen(specimen);
        guiLabResult.setSubscript(subscript);
        addSiteDataToGUIBean(guiLabResult, labResult);
        guiLabResult.setPatientName(patName);
        guiLabResult.setAccessionType(accessionType);
        guiLabResult.setPerformingLabName(performingLabName);
        guiLabResult.setPerformingLabStreet1(performingLabAddress);
        guiLabResult.setRequestingLocation(requestingLocation);


        if (labResult.getCdrEventId() == null || labResult.getCdrEventId().length() == 0) {
            guiLabResult.setEventId(labResult.getOrderId());
        } else {
            guiLabResult.setEventId(labResult.getCdrEventId());
        }
        guiLabResult.setSensitive(labResult.isSensitive());
        if (guiLabResult.isSensitive()) {
            guiLabResult.setTestName(guiLabResult.getTestName() + "  ** Sensitive **");
        }

        List<GUICode> guiCodes = new ArrayList<GUICode>();
        for (Code code : labResult.getCodes()) {
            guiCodes.add(dozerBeanMapper.map(code, GUICode.class));
        }
        guiLabResult.setCodes(guiCodes);

        return guiLabResult;
    }

    public GUILabGraph getLabGraphData(GUIUser user, GUIPatient patient, String itemId, Calendar startDate, Calendar endDate, String recordSiteCode) {
        try {
            LabResult[] labResults = getServiceHandler().getPatientLabTestResults(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(startDate)
                            .endDate(endDate)
                            .itemId(itemId)
                            .recordSiteCode(recordSiteCode)
                            .build());

            return toGUILabGraph(labResults);

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    private GUILabGraph toGUILabGraph(LabResult[] labResults) {
        XMLGregorianCalendar date;
        double result = 0;
        XMLGregorianCalendar vD;
        String vR = "";

        int count = 0;
        double total = 0;
        double min = 0;
        double max = 0;
        double avg = 0;
        //double hFactor = 0;
        XMLGregorianCalendar mindate = null;
        XMLGregorianCalendar maxdate = null;

        List<GUILabResult> guiLabResults = new ArrayList<GUILabResult>();
        GUILabGraph guiLabGraph = new GUILabGraph();

        for (LabResult res : labResults) {
            try {
                vD = res.getResultDate();
                vR = res.getResult();

                if (vD == null || vR == null) continue;

                result = Double.parseDouble(vR);
                guiLabResults.add(toGUILabResult(res));
            } catch (NumberFormatException e) {
                //do not include non-numeric results in lab graph result set
                guiLabGraph.setExceptionFlag(true);
                if (logger.isDebugEnabled()) {
                    logger.debug("Bad value in lab result cell: '" + vR + "'");
                }
                continue;
            }
            date = vD;
            count++;
            total = total + result;
            if (count == 1) {
                min = result;
                max = result;
                mindate = date;
                maxdate = date;
            } else {
                if (min > result) {
                    min = result;
                    mindate = date;
                }
                if (max < result) {
                    max = result;
                    maxdate = date;
                }
            }
        }

        guiLabGraph.setGuiLabResults(guiLabResults);

        if (count > 0) avg = total / count;

        guiLabGraph.setCount(count);
        guiLabGraph.setAvg(roundAverage(avg));
        guiLabGraph.setTotal(total);
        guiLabGraph.setMax(max);
        guiLabGraph.setMin(min);
        guiLabGraph.setMinDate(Utils.formatDisplayDate(mindate));
        guiLabGraph.setMaxDate(Utils.formatDisplayDate(maxdate));

        return guiLabGraph;
    }

    private double roundAverage(Double avg) {
        String s = avg.toString();
        int i = s.indexOf('.');
        if (i >= 0) {
            int o = 0;
            if (s.length() > i + 3) {
                o = Integer.parseInt(s.substring(i + 3, i + 4));
                s = s.substring(0, i + 3);
            }
            avg = Double.valueOf(s);
            if (o > 4) avg += .01;
        }
        return avg;
    }

    public List<GUILabResult> getProviderLabAbnormalResults(
            GUIUser user,
            Calendar startDate,
            Calendar endDate,
            byte[] encryptSecurityKey) {
        try {
            LabResult[] labResults = getServiceHandler().getProviderLabAbnormalResults(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());

            List<GUILabResult> guiLabResults = new ArrayList<GUILabResult>();

            for (LabResult labResult : labResults) {
                GUILabResult lr = toGUILabResult(labResult);
                lr.setPatientToken(getProvQueryPatientToken(encryptSecurityKey, labResult));
                guiLabResults.add(lr);
            }

            return guiLabResults;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public List<GUILabResult> getPatientLabResults(
            GUIUser user,
            GUIPatient patient,
            Calendar startDate,
            Calendar endDate) {
        try {
            LabResult[] labResults = getServiceHandler().getPatientLabResults(

                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());

            List<GUILabResult> guiLabResults = new ArrayList<GUILabResult>();

            for (LabResult labResult : labResults) {
                GUILabResult lr = toGUILabResult(labResult);
                guiLabResults.add(lr);
            }

            return guiLabResults;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }
}
