/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package gov.va.med.dao;

import gov.va.med.enterprise.bean.ResponseGUIPatientQuery;
import gov.va.med.enterprise.bean.ResponseGUIPatientSelect;
import gov.va.med.foundations.utilities.EncryptUtil;
import gov.va.med.guibeans.*;
import gov.va.med.jmeadows.webservice.*;
import gov.va.med.vhahon.common.Utils;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import javax.xml.soap.SOAPException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 *
 *
 */
@Component
public class PatientDao extends DaoBase {
    private static final Logger logger = Logger.getLogger(PatientDao.class.getName());

    public ResponseGUIPatientQuery lookupPatientPDWS(GUIUser user, PdwsQueryBean queryBean, byte[] encryptSecurityKey) {
        try {
            queryBean.setUser(dozerBeanMapper.map(user, User.class));
            queryBean.setRequestingApp(appConfig.getAppName());

            ResponsePatientQuery respPatQuery = getServiceHandler().lookupPatientPDWS(queryBean);
            List<Patient> patients = respPatQuery.getPatients();

            ResponseGUIPatientQuery respGUIPatQuery = new ResponseGUIPatientQuery();
            List<GUILookupPatient> guiLookupPatients = new ArrayList<GUILookupPatient>();

            // Going to compare user's SSN with patient's SSN to prevent user from viewing their own medical records
            String userSSN = EncryptUtil.decrypt(user.getEncryptedFederatedUid());
            String dob;
            String gender;

            for (Patient patient : patients) {
                String fmpssn = patient.getSSN();
                fmpssn = Utils.formattedSSN(fmpssn);
                dob = patient.getDob();
                if (dob == null || dob.isEmpty()) {
                    dob = "N/A";
                }
                gender = patient.getGender();
                if (gender == null || gender.isEmpty()) {
                    gender = "N/A";
                }

                String patientToken = generatePatientToken(encryptSecurityKey, appConfig.getIdTypeEDIPI(), patient.getEDIPI());

                GUILookupPatient guiLookupPatient = new GUILookupPatient();
                guiLookupPatient.setName(patient.getName());
                guiLookupPatient.setEdipi(patient.getEDIPI());
                guiLookupPatient.setDob(dob);
                guiLookupPatient.setFmpssn(fmpssn);
                guiLookupPatient.setSponsorSsn(patient.getSponsorSSN());
                guiLookupPatient.setGender(gender);
                guiLookupPatient.setSensitive(patient.isSensitive());
                guiLookupPatient.setPatientToken(patientToken);

                // User is VA
                if ("VA".equalsIgnoreCase(user.getAgency())) {
                    // User is in list of searched patients
                    if (userSSN.equalsIgnoreCase(fmpssn)) {
                        // Clear patientToken so that this patient (self) cannot be selected (link in table will be disabled)
                        guiLookupPatient.setPatientToken("selfselectaccessdenied");
                    }
                }

                guiLookupPatients.add(guiLookupPatient);
            }

            respGUIPatQuery.setSuccess(respPatQuery.isSuccess());
            respGUIPatQuery.setErrorMsg(respPatQuery.getErrorMsg());
            respGUIPatQuery.setGuiPatients(guiLookupPatients);

            return respGUIPatQuery;
        } catch (SOAPException ex) {
            throw new DaoException(ex);
        }
    }

    public ResponseGUIPatientSelect selectPatient(GUIUser user, Patient patient, String patientId, String idType, String patientSiteCode) {
        try {

            JMeadowsQuery queryBean =
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .itemId(patientId)
                            .status(idType)
                            .recordSiteCode(patientSiteCode)
                            .build();

            // Selected patient's info to be stored in the user's recently viewed patients table
            if (patient != null) {
                queryBean.setPatient(patient);
            }

            ResponsePatientSelect response = getServiceHandler().selectPatientMVI(queryBean);

            ResponseGUIPatientSelect guiResponse = new ResponseGUIPatientSelect();

            guiResponse.setSuccess(response.isSuccess());
            guiResponse.setVARestricted(response.isVARestricted());
            guiResponse.setAllowVAAccess(response.isAllowVAAccess());
            guiResponse.setErrorMsg(response.getErrorMsg());
            guiResponse.setGUIPatient(dozerBeanMapper.map(response.getPatient(), GUIPatient.class));

            List<GUIExternalID> externalIDs = new ArrayList<GUIExternalID>();
            for (ExternalID eid : response.getExternalIDs()) {
                externalIDs.add(dozerBeanMapper.map(eid, GUIExternalID.class));
            }

            guiResponse.setExternalIDs(externalIDs);

            return guiResponse;
        } catch (SOAPException ex) {
            throw new DaoException(ex);
        }
    }

    public void selectPatientForVASensitive(GUIUser user, GUIPatient patient) {
        try {
            getServiceHandler().selectPatientForVASensitive(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .build());
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public void auditVARestrictedAccess(GUIUser user, GUIPatient patient, String auditText) {
        try {
            JMeadowsQuery queryBean =
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .build();

            queryBean.setItemId(auditText);

            getServiceHandler().auditVARestrictedAccess(queryBean);
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public List<GUIPatientDemographics> getPatientDemographics(GUIUser user, GUIPatient patient, boolean includeVLER) {
        try {
            List<PatientDemographics> patDemos = getServiceHandler().getPatientDemographics(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .includeVLER(includeVLER)
                            .build());

            List<GUIPatientDemographics> guiPatDemos = new ArrayList<GUIPatientDemographics>();

            for (PatientDemographics patDemo : patDemos) {
                String name = "";
                String address1 = "";
                String address2 = "";
                String city = "";
                String state = "";
                String zipCode = "";
                String fmpssn = "";
                String edipi = "";
                String pcm = "";
                String pcmPhone = "";
                String pcmDate = "";
                String pcmClinic = "";
                String pcTeam = "";
                String age = "";
                String dob = "";
                String gender = "";
                String race = "";
                String phone1 = "";
                String phone2 = "";
                boolean isCodeGray = false;
                String codeGrayMsg = "";
                String percentServiceConnected = "";
                String clinic = "";
                String rank = "";
                String fmp = "";
                String sponsorName = "";
                String languages = "";
                String religion = "";
                String maritalStatus = "";


                if (patDemo.getName() != null) {
                    name = patDemo.getName();
                }

                if (patDemo.getAddress1() != null) {
                    address1 = patDemo.getAddress1();
                }

                if (patDemo.getAddress2() != null) {
                    address2 = patDemo.getAddress2();
                }

                if (patDemo.getCity() != null) {
                    city = patDemo.getCity();
                }

                if (patDemo.getState() != null) {
                    state = patDemo.getState();
                }

                if (patDemo.getZipCode() != null) {
                    zipCode = patDemo.getZipCode();
                }

                if (patDemo.getSSN() != null) {
                    fmpssn = patDemo.getSSN();
                }

                if (patDemo.getPrimaryProvider() != null) {
                    pcm = patDemo.getPrimaryProvider();
                }

                if (patDemo.getPcmPhone() != null) {
                    pcmPhone = patDemo.getPcmPhone();
                }

                if (patDemo.getPcmDates() != null) {
                    pcmDate = Utils.formatDisplayDate(patDemo.getPcmDates());
                }

                if (patDemo.getPcmClinic() != null) {
                    pcmClinic = patDemo.getPcmClinic();
                }

                if (patDemo.getPcTeam() != null) {
                    pcTeam = patDemo.getPcTeam();
                }

                if (patDemo.getAge() != null) {
                    age = patDemo.getAge();
                }

                if (patDemo.getDob() != null) {
                    dob = patDemo.getDob();
                }

                if (patDemo.getGender() != null) {
                    gender = patDemo.getGender();
                }

                if (patDemo.getRace() != null && !patDemo.getRace().isEmpty()) {
                    race = patDemo.getRace().get(0).getDisplay();
                }

                if (patDemo.getPhone1() != null) {
                    phone1 = patDemo.getPhone1();
                }

                if (patDemo.getPhone2() != null) {
                    phone2 = patDemo.getPhone2();
                }

                if (patDemo.getCodeGreen() != null) {
                    if (!patDemo.getCodeGreen().equals("0") && patDemo.getCodeGreen().length() > 0) {
                        isCodeGray = true;
                        codeGrayMsg = patDemo.getCodeGreen();
                    }
                }

                if (patDemo.getPercentServiceConnected() != null) {
                    percentServiceConnected = patDemo.getPercentServiceConnected();
                }

                if (patDemo.getClinic() != null) {
                    clinic = patDemo.getClinic();
                }

                if (patDemo.getFMP() != null) {
                    fmp = patDemo.getFMP();
                }

                if (patDemo.getEDIPI() != null) {
                    edipi = patDemo.getEDIPI();
                }

                if (patDemo.getRank() != null) {
                    rank = patDemo.getRank();
                }

                if (patDemo.getSponsorName() != null) {
                    sponsorName = patDemo.getSponsorName();
                }
//
//                String languages = "";
//                String religion = "";
//                String maritalStatus = "";

                if(patDemo.getLanguages() != null ){
                    StringBuilder langSb = new StringBuilder();
                    int count = 0;
                    for(String language : patDemo.getLanguages()){
                        if(count > 0) {
                            langSb.append("; ");
                        }
                        langSb.append(language);
                        count++;

                    }
                    languages = langSb.toString();
                }

                if(patDemo.getReligion() != null){
                    StringBuilder relSb = new StringBuilder();
                    int count = 0;
                    for(String rel : patDemo.getReligion()){
                        if(count > 0) {
                            relSb.append("; ");
                        }
                        relSb.append(rel);
                        count++;

                    }
                    religion = relSb.toString();
                }

                List<GUIPerson> emergencyContacts = new ArrayList<GUIPerson>();
                if(patDemo.getEmergencyContact() != null && patDemo.getEmergencyContact().size() > 0){
                    for(Person person : patDemo.getEmergencyContact()){
                        emergencyContacts.add(dozerBeanMapper.map(person, GUIPerson.class));
                    }
                }

                List<GUIPerson> nextOfKin = new ArrayList<GUIPerson>();
                if(patDemo.getNextOfKin() != null && patDemo.getNextOfKin().size() > 0){
                    for(Person person : patDemo.getNextOfKin()){
                        nextOfKin.add(dozerBeanMapper.map(person, GUIPerson.class));
                    }
                }

                GUIPatientDemographics guiPatDemo = new GUIPatientDemographics();
                guiPatDemo.setName(name);
                guiPatDemo.setAddress1(address1);
                guiPatDemo.setAddress2(address2);
                guiPatDemo.setCity(city);
                guiPatDemo.setState(state);
                guiPatDemo.setZipCode(zipCode);
                guiPatDemo.setFmpssn(fmpssn);
                guiPatDemo.setPcm(pcm);
                guiPatDemo.setPcmPhone(pcmPhone);
                guiPatDemo.setPcmDate(pcmDate);
                guiPatDemo.setPcmClinic(pcmClinic);
                guiPatDemo.setPcTeam(pcTeam);
                guiPatDemo.setAge(age);
                guiPatDemo.setDob(dob);
                guiPatDemo.setGender(gender);
                guiPatDemo.setRace(race);
                guiPatDemo.setPhone1(phone1);
                guiPatDemo.setPhone2(phone2);
                guiPatDemo.setCodeGray(isCodeGray);
                guiPatDemo.setCodeGrayMessage(codeGrayMsg);
                guiPatDemo.setPercentServiceConnected(percentServiceConnected);
                guiPatDemo.setClinic(clinic);
                guiPatDemo.setEdipi(edipi);

                guiPatDemo.setReligion(religion);
                guiPatDemo.setMaritalStatus(patDemo.getMaritalStatus());
                guiPatDemo.setLanguages(languages);

                guiPatDemo.setEmergencyContact(emergencyContacts);
                guiPatDemo.setNextOfKin(nextOfKin);
                guiPatDemo.setRank(rank);
                guiPatDemo.setFmp(fmp);
                guiPatDemo.setSponsorName(sponsorName);

                List<GUICode> guiCodes = new ArrayList<GUICode>();
                for (Code code : patDemo.getRace()) {
                    guiCodes.add(dozerBeanMapper.map(code, GUICode.class));
                }
                guiPatDemo.setCodes(guiCodes);

                addSiteDataToGUIBean(guiPatDemo, patDemo);
                guiPatDemos.add(guiPatDemo);


            }

            //sort by site moniker
            Collections.sort(guiPatDemos, new SiteAlphaComparator());

            //push lookup site to the top of the list
            String hostSiteCode = user.getHostSite().getSiteCode();

            int localSiteIdx = -1;
            int demosSize = guiPatDemos.size();
            for (int i = 0; i < demosSize; i++) {
                GUIPatientDemographics guiPatDemo = guiPatDemos.get(i);
                if (guiPatDemo.getSite().getSiteCode().equals(hostSiteCode)) {
                    localSiteIdx = i;
                    break;
                }
            }

            if (localSiteIdx > -1) {
                GUIPatientDemographics guiPatDemo = guiPatDemos.get(localSiteIdx);
                guiPatDemos.remove(localSiteIdx);
                guiPatDemos.add(0, guiPatDemo);
            }

            return guiPatDemos;
        } catch (SOAPException ex) {
            throw new DaoException(ex);
        }
    }

    public List<GUIPatientDemographicDetail> getPatientDemographicDetails(GUIUser user, GUIPatient patient) {
        try {
            List<PatientDemographicsDetail> patDemoDetails = getServiceHandler()
                    .getAllPatientDemographicDetails(
                            jMeadowsQueryBuilderFactory.getBuilder()
                                    .user(user)
                                    .patient(patient)
                                    .build());
            List<GUIPatientDemographicDetail> guiPatDemoDetails = new ArrayList<GUIPatientDemographicDetail>();

            for (PatientDemographicsDetail patDemoDetail : patDemoDetails) {
                GUIPatientDemographicDetail guiDemoDetail = new GUIPatientDemographicDetail();
                guiDemoDetail.setReport(patDemoDetail.getReport());

                addSiteDataToGUIBean(guiDemoDetail, patDemoDetail);

                guiPatDemoDetails.add(guiDemoDetail);
            }

            //sort by site moniker
            Collections.sort(guiPatDemoDetails, new SiteAlphaComparator());

            //push lookup site to the top of the list
            String hostSiteCode = user.getHostSite().getSiteCode();

            int localSiteIdx = -1;
            int detailsSize = guiPatDemoDetails.size();
            for (int i = 0; i < detailsSize; i++) {
                GUIPatientDemographicDetail guiPatDemo = guiPatDemoDetails.get(i);
                if (guiPatDemo.getSite().getSiteCode().equals(hostSiteCode)) {
                    localSiteIdx = i;
                    break;
                }
            }

            if (localSiteIdx > -1) {
                GUIPatientDemographicDetail guiPatDemo = guiPatDemoDetails.get(localSiteIdx);
                guiPatDemoDetails.remove(localSiteIdx);
                guiPatDemoDetails.add(0, guiPatDemo);
            }

            return guiPatDemoDetails;
        } catch (SOAPException ex) {
            throw new DaoException(ex);
        }
    }
}
