/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package gov.va.med.dao;

import gov.va.med.foundations.utilities.EncryptUtil;
import gov.va.med.guibeans.*;
import gov.va.med.jmeadows.webservice.Code;
import gov.va.med.jmeadows.webservice.Endpoint;
import gov.va.med.jmeadows.webservice.RadiologyReport;
import gov.va.med.vhahon.common.Utils;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

@Component
public class RadiologyDao extends DaoBase {
    private static final Logger logger = Logger.getLogger(RadiologyDao.class.getName());

    public List<GUIRadiologyReport> getPatientRads(GUIUser user, GUIPatient patient, Calendar startDate, Calendar endDate) {
        try {
            RadiologyReport[] rads = getServiceHandler().getPatientRads(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());

            List<GUIRadiologyReport> guiRads = new ArrayList<GUIRadiologyReport>();

            for (int i = 0; i < rads.length; i++) {
                RadiologyReport rad = rads[i];

                String examDate = "";
                String study = "";
                String status = "";
                String examId = "";
                String accessionNumber = "";
                String orderNumber = "";
                String priority = "";
                String amendedDate = "";

                if (rad.getExamDate() != null) {
                    examDate = Utils.formatDisplayDate(rad.getExamDate());
                }

                if (rad.getExamDate() != null) {
                    amendedDate = Utils.formatDisplayDate(rad.getAmendedDate());
                }

                if (rad.getStudy() != null) {
                    study = rad.getStudy();
                }

                if (rad.getStatus() != null) {
                    status = rad.getStatus();
                }

                if (rad.getExamId() != null) {
                    examId = rad.getExamId();
                }

                if (rad.getAccessionNumber() != null) {
                    accessionNumber = rad.getAccessionNumber();
                }

                if (rad.getOrderNumber() != null) {
                    orderNumber = rad.getOrderNumber();
                }

                if (rad.getPriority() != null) {
                    priority = rad.getPriority();
                }

                GUIRadiologyReport guiRad = new GUIRadiologyReport();

                if (rad.getSite() != null) {
                    if (rad.getSite().getEndpoints() != null) {
                        //set medweb available if medweb endpoint exists
                        for (Endpoint endpoint : rad.getSite().getEndpoints()) {
//                            if (appConfig.getProtocolMedweb().equalsIgnoreCase(endpoint.getProtocol()))
                            if (endpoint.getProtocol().startsWith(appConfig.getProtocolMedweb())) {
                                guiRad.setMedWebAvailable(true);
                                break;
                            }
                        }
                    }
                }

                guiRad.setAccessionNumber(accessionNumber);
                guiRad.setExamId(examId);
                guiRad.setExamDate(examDate);
                guiRad.setStatus(status);
                guiRad.setStudy(study);
                guiRad.setOrderNumber(orderNumber);
                guiRad.setPriority(priority);
                guiRad.setAmendedDate(amendedDate);

                mapToGUIRadiologyReport(rad, guiRad);
                //result includes radiology report details
                if (StringUtils.isNotBlank(rad.getReportText())) {
                  //guiRad.setExamId(appConfig.getCachedTag() + i);
                    guiRad.setSessionCachedRecord(true);
                }
                guiRad.setCacheId(appConfig.getCachedTag() + i);

                List<GUICode> guiRadCodes = new ArrayList<GUICode>();
                for (Code radCode : rad.getCodes()) {
                    guiRadCodes.add(dozerBeanMapper.map(radCode, GUICode.class));
                }
                guiRad.setCodes(guiRadCodes);

                addSiteDataToGUIBean(guiRad, rad);
                addMedWebEndpointToSite(guiRad, rad);

                guiRads.add(guiRad);
            }

            return guiRads;
        } catch (RuntimeException e) {
            throw new DaoException(e);
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    private void addMedWebEndpointToSite(GUIRadiologyReport guiRad, RadiologyReport rad) {
        List<GUIEndpoint> guiEndpoints = new ArrayList<GUIEndpoint>();
        if (rad.getSite() != null) {
            for (Endpoint endpoint : rad.getSite().getEndpoints()) {
                if (endpoint.getProtocol() != null && endpoint.getProtocol().startsWith(appConfig.getProtocolMedweb())) {
                    guiEndpoints.add(dozerBeanMapper.map(endpoint, GUIEndpoint.class));
                }
            }
            if (guiRad.getSite() != null && !guiEndpoints.isEmpty()) {
                guiRad.getSite().setEndpoints(guiEndpoints);
            }
        }
    }

    private void mapToGUIRadiologyReport(RadiologyReport radReport, GUIRadiologyReport guiRadReport) {
        String accessionNumber = "";
        String priority = "";
        String procedure = "";
        String examDate = "";
        String amendedDate = "";
        String orderReason = "";
        String resultCode = "";
        String interpretingHCP = "";
        String approvingHCP = "";
        String transcribeDate = "";
        String reportStatus = "";
        String reportText = "";
        String impressionText = "";
        String abnormal = "";

        if (radReport.getAccessionNumber() != null) {
            accessionNumber = radReport.getAccessionNumber();
        }

        if (radReport.getReportText() != null) {
            reportText = radReport.getReportText();
            reportText = Utils.stringReplaceChars(reportText, (char) 10, "<br>");
        }

        if (radReport.getPriority() != null) {
            priority = radReport.getPriority();
        }

        if (radReport.getStudy() != null) {
            procedure = radReport.getStudy();
        }

        if (radReport.getExamDate() != null) {
            examDate = Utils.formatDisplayDate(radReport.getExamDate());
        }

        if (radReport.getAmendedDate() != null) {
            amendedDate = Utils.formatDisplayDate(radReport.getAmendedDate());
        }

        if (radReport.getReasonForOrder() != null) {
            orderReason = radReport.getReasonForOrder();
            orderReason = Utils.stringReplaceChars(orderReason, (char) 10, "<br>");
        }

        if (radReport.getResultCode() != null) {
            resultCode = radReport.getResultCode();
        }

        if (radReport.getAbnormal() != null) {
            abnormal = radReport.getAbnormal();
        }

        if (radReport.getInterpretingHCP() != null) {
            interpretingHCP = radReport.getInterpretingHCP();
        }

        if (radReport.getApprovedBy() != null) {
            approvingHCP = radReport.getApprovedBy();
        }

        if (radReport.getTranscribeDate() != null) {
            transcribeDate = Utils.formatDisplayDate(radReport.getTranscribeDate());
        }

        if (radReport.getReportStatus() != null) {
            reportStatus = radReport.getReportStatus();
        }

        if (radReport.getImpressionText() != null) {
            impressionText = radReport.getImpressionText();
            impressionText = Utils.stringReplaceChars(impressionText, (char) 10, "<br>");
        }

        //display a plain text report for the vista rpc & dod adapter protocols
        if (appConfig.getProtocolRPC().equalsIgnoreCase(radReport.getSourceProtocol())) {
            guiRadReport.setPlainTextReport(true);
        }

        guiRadReport.setAccessionNumber(accessionNumber);
        guiRadReport.setProcedure(procedure);
        guiRadReport.setPriority(priority);
        guiRadReport.setExamDate(examDate);
        guiRadReport.setOrderReason(orderReason);
        guiRadReport.setResultCode(resultCode);
        guiRadReport.setAbnormal(abnormal);
        guiRadReport.setOrderReason(orderReason);
        guiRadReport.setInterpretingHCP(interpretingHCP);
        guiRadReport.setApprovingHCP(approvingHCP);
        guiRadReport.setTranscribeDate(transcribeDate);
        guiRadReport.setImpressionText(impressionText);
        guiRadReport.setReportStatus(reportStatus);
        guiRadReport.setReportText(reportText);
        guiRadReport.setAmendedDate(amendedDate);
    }

    public GUIRadiologyReport getRadiologyReport(GUIUser user, GUIPatient patient, String radiologyExamId, String recordSiteCode) {
        try {
            RadiologyReport radReport = getServiceHandler().getRadiologyReport(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .itemId(radiologyExamId)
                            .recordSiteCode(recordSiteCode)
                            .build());

            GUIRadiologyReport guiReport = new GUIRadiologyReport();

            mapToGUIRadiologyReport(radReport, guiReport);

            if (radReport.getSite() != null) {
                if (radReport.getSite().getEndpoints() != null) {
                    //set medweb available if medweb endpoint exists
                    for (Endpoint endpoint : radReport.getSite().getEndpoints()) {
//                        if (appConfig.getProtocolMedweb().equalsIgnoreCase(endpoint.getProtocol()))
                        if (endpoint.getProtocol().startsWith(appConfig.getProtocolMedweb())) {
                            guiReport.setMedWebAvailable(true);
                            break;
                        }
                    }
                }
            }

            addSiteDataToGUIBean(guiReport, radReport);

            return guiReport;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public GUIMedWebAccess getMedWebAccessBean(GUIUser user, GUIPatient patient,
                                               String accessionNumber, String medWebURL,
                                               String medWebUsrEnc, String medWebPwdEnc,
                                               String aeTitle, String accessionPrefix) {

        GUIMedWebAccess guiMedWeb = new GUIMedWebAccess();

        if (StringUtils.isNotBlank(accessionPrefix)) {
            accessionNumber = accessionPrefix + accessionNumber;
        }

        guiMedWeb.setAccessionNumber(accessionNumber);
        guiMedWeb.setAeTitle(aeTitle);

        guiMedWeb.setMedWebUrl(medWebURL);

        //attempt to obfuscate medweb credentials
        byte[] decryptedUsrEnc;
        byte[] decryptedPwdEnc;

        try {
            decryptedUsrEnc = EncryptUtil.decrypt(medWebUsrEnc).getBytes("UTF-8");
            decryptedPwdEnc = EncryptUtil.decrypt(medWebPwdEnc).getBytes("UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new DaoException(e);
        }
        String base64UserName = Base64.encodeBase64URLSafeString(decryptedUsrEnc);
        String base64Password = Base64.encodeBase64URLSafeString(decryptedPwdEnc);

        guiMedWeb.setMedWebUserName(base64UserName);
        guiMedWeb.setMedWebPassword(base64Password);

        try {
            getServiceHandler().logAudit(user.getHostSite().getSiteCode(), user.getUserIen(),
                    user.getProviderIen(), "", user.getName(),
                    patient.getEDIPI() + "^" + aeTitle + "^" + accessionNumber, "Medweb", "", "");
        } catch (Exception e) {
            throw new DaoException(e);
        }

        return guiMedWeb;
    }
}
