/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package gov.va.med.dao;

import gov.va.med.foundations.utilities.EncryptUtil;
import gov.va.med.foundations.utilities.VistaKernelHash;
import gov.va.med.guibeans.*;
import gov.va.med.jmeadows.webservice.*;
import gov.va.med.vhahon.common.Utils;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.util.*;

/**
 *
 *
 */
@Component
public class UserDao extends DaoBase {
    private static final Logger logger = Logger.getLogger(UserDao.class.getName());

    public GUIUser login(String cardID, String agency, String siteCode, String accessCode, String verifyCode, String ipAddress, String subjectDN) {
        try {
            if ("VA".equalsIgnoreCase(agency)) {
                accessCode = VistaKernelHash.encrypt(accessCode, false);
                verifyCode = VistaKernelHash.encrypt(verifyCode, false);
            }

            User user = getServiceHandler().loginEnterprise(cardID, siteCode, accessCode, verifyCode, ipAddress, subjectDN);

            return dozerBeanMapper.map(user, GUIUser.class);

        } catch (Exception ex) {
            throw new DaoException(ex);
        }
    }

    public String getJanusGUIConfig(GUIUser user) {
        try {
            return getServiceHandler().getJanusGUIConfig(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .itemId("")
                            .build());
        } catch (Exception ex) {
            throw new DaoException(ex);
        }
    }

    public boolean setJanusGUIConfig(GUIUser user, String cfg) {
        try {
            return getServiceHandler().setJanusGUIConfig(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .itemId("")
                            .build(), cfg);
        } catch (Exception ex) {
            throw new DaoException(ex);
        }
    }

    public IehrUserProfile getIehrUserProfile(String smartCardID, String smartCardAgency) {
        try {
            return getServiceHandler().getIehrUserProfile(smartCardID, smartCardAgency);
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public boolean setIehrUserProfile(IehrUserProfile userProfile) {
        try {
            return getServiceHandler().setIehrUserProfile(userProfile);
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public List<GUISite> getSites() {
        try {

            List<Site> sites = getServiceHandler().getSites();
            List<GUISite> guiSites = null;
            if (sites != null) {
                guiSites = new ArrayList<GUISite>(sites.size());
                for (Site site : sites) {
                    GUISite guiSite = dozerBeanMapper.map(site, GUISite.class);
                    guiSites.add(guiSite);
                }
            }
            return guiSites;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public List<GUIAuditEntry> getLoginInfo(String smartCardID) {
        try {
            List<AuditEntry> auditEntries = getServiceHandler().getLoginInfo(smartCardID);

            List<GUIAuditEntry> guiAuditEntries = new ArrayList<GUIAuditEntry>();

            for (AuditEntry aEntry : auditEntries) {
                GUIAuditEntry guiAuditEntry = new GUIAuditEntry();

                String auditID = "";
                String entryDate = "";
                String startDate = "";
                String endDate = "";
                String systemID = "";
                String userID = "";
                String userNPI = "";
                String userName = "";
                String patID = "";
                String category = "";
                String queryType = "";
                String cardID = "";
                String ipAddress = "";

                if (aEntry.getAuditID() != null) {
                    auditID = aEntry.getAuditID();
                }

                if (aEntry.getEntryDate() != null) {
                    entryDate = Utils.formatDisplayDateTime(aEntry.getEntryDate());
                }

                if (aEntry.getIpAddress() != null) {
                    ipAddress = aEntry.getIpAddress();
                }

                if (aEntry.getUserID() != null) {
                    userID = aEntry.getUserID();
                }

                guiAuditEntry.setAuditID(auditID);
                guiAuditEntry.setEntryDate(entryDate);
                guiAuditEntry.setStartDate(startDate);
                guiAuditEntry.setEndDate(endDate);
                guiAuditEntry.setSystemID(systemID);
                guiAuditEntry.setUserID(userID);
                guiAuditEntry.setUserNPI(userNPI);
                guiAuditEntry.setUserName(userName);
                guiAuditEntry.setPatID(patID);
                guiAuditEntry.setCategory(category);
                guiAuditEntry.setQueryType(queryType);
                guiAuditEntry.setCardID(cardID);
                guiAuditEntry.setIpAddress(ipAddress);

                guiAuditEntries.add(guiAuditEntry);
            }

            return guiAuditEntries;

        } catch (Exception ex) {
            throw new DaoException(ex);
        }
    }

    public List<GUILookupPatient> getRecentlyViewedPatients(GUIUser user, byte[] securityKey)
    {
        try {
            List<Patient> patients = getServiceHandler().getRecentlyViewedPatients(user.getUserId());
            List<GUILookupPatient> guiPatients = new ArrayList<GUILookupPatient>();
            GUILookupPatient guiPatient;

            for (Patient patient : patients) {
                guiPatient = new GUILookupPatient();

                // Name
                guiPatient.setName(patient.getName());
                // EDIPI
                guiPatient.setEdipi(patient.getEDIPI());
                // SSN
                guiPatient.setFmpssn(patient.getSSN());
                // Sponsor SSN
                guiPatient.setSponsorSsn(patient.getFMP());
                // Gender
                guiPatient.setGender(patient.getGender());
                // DoB
                guiPatient.setDob(patient.getDob());
                // Sensitive
                guiPatient.setSensitive(patient.isSensitive());
                // Patient Token
                guiPatient.setPatientToken(generatePatientToken(securityKey, "EDIPI", patient.getEDIPI()));

                guiPatients.add(guiPatient);
            }

            return guiPatients;
        }
        catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public GUIOutsideProvider loginOutsideProvider(String userName, String hashedPassword)
    {
        GUIOutsideProvider guiOutsideProvider = null;

        try {
            OutsideProvider outsideProvider = getServiceHandler().getOutsideProvider(null, userName, hashedPassword);

            if (outsideProvider != null) {
                guiOutsideProvider = dozerBeanMapper.map(outsideProvider, GUIOutsideProvider.class);
            }
        }
        catch (Exception e) {
            throw new DaoException(e);
        }

        return guiOutsideProvider;
    }
}
