/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */
 
 
package gov.va.med.rpc;


import gov.va.med.foundations.utilities.ARMLog;
import gov.va.med.vds.vistabeans.Site;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class RPCConnectionManager {
    private static final Logger LOGGER = LoggerFactory.getLogger(RPCConnectionManager.class);

    protected LinkedHashMap<Site, RPCConnection> cxnTbl;
    protected String modality;
    private long threadTimeoutMS = 30000;
    private int maxThreadPoolSize = 10;

    public RPCConnectionManager(LinkedHashMap<Site, RPCConnection> cxnTbl) {
        if (cxnTbl == null) {
            throw new IllegalArgumentException("RPCConnection Table is null");
        }

        this.cxnTbl = cxnTbl;
        Collection<RPCConnection> values = this.cxnTbl.values();
        if (values.iterator().hasNext()) {
            this.modality = ((RPCConnection) values.iterator().next()).getRPCEndpoint().getModality();
        }
    }

    public RPCConnectionManager(RPCConnection cxn) {
        if (cxn == null) {
            throw new IllegalArgumentException("RPCConnection is null");
        }

        this.modality = cxn.getRPCEndpoint().getModality();
        cxnTbl = new LinkedHashMap<Site, RPCConnection>();
        addConnection(cxn);
    }

    public RPCConnectionManager(Site site) {
        if (site == null) {
            throw new IllegalArgumentException("Site is null");
        }

        RPCConnection cxn = new RPCConnection(site);
        this.modality = cxn.getRPCEndpoint().getModality();
        cxnTbl = new LinkedHashMap<Site, RPCConnection>();
        addConnection(cxn);
    }

    public RPCConnectionManager(List<Site> sites) {
        cxnTbl = new LinkedHashMap<Site, RPCConnection>();
        for (Site site : sites) {
            RPCConnection cxn = new RPCConnection(site);
            if (StringUtils.isEmpty(this.modality)) {
                this.modality = cxn.getRPCEndpoint().getModality();
            }
            addConnection(cxn);
        }
    }

    public boolean siteIsConnected(Site site) {
        return cxnTbl.containsKey(site);

    }

    public long getThreadTimeoutMS() {
        return threadTimeoutMS;
    }

    public void setThreadTimeoutMS(long threadTimeoutMS) {
        this.threadTimeoutMS = threadTimeoutMS;
    }

    public int getMaxThreadPoolSize() {
        return maxThreadPoolSize;
    }

    public void setMaxThreadPoolSize(int maxThreadPoolSize) {
        this.maxThreadPoolSize = maxThreadPoolSize;
    }

    private void checkModality(RPCConnection cxn) {
        checkModality(cxn.getRPCEndpoint().getModality());
    }

    private void checkModality(String modality) {
        if (!modality.equals(this.modality)) {
            throw new RuntimeException("Invalid modality");
        }
    }

    public final void addConnection(RPCConnection cxn) {
        if (cxnTbl == null) {
            cxnTbl = new LinkedHashMap<Site, RPCConnection>();
            this.modality = cxn.getRPCEndpoint().getModality();
        }
        checkModality(cxn);
        if (!cxnTbl.containsKey(cxn.getSite())) {
            cxnTbl.put(cxn.getSite(), cxn);
        }
    }

    public void removeConnection(Site site) throws RPCException {
        if (cxnTbl.containsKey(site)) {
            RPCConnection cxn = cxnTbl.get(site);
            if (cxn.isConnected()) {
                cxn.disconnect();
            }
            cxnTbl.remove(site);
        }
    }

    public LinkedHashMap<Site, RPCConnection> getConnectionTable() {
        return cxnTbl;
    }

    public void addConnections(List<Site> sites) throws RPCException {
        for (Site site : sites) {
            addConnection(new RPCConnection(site));
        }
    }

    public void addConnections(LinkedHashMap<Site, RPCConnection> connectionTable) throws RPCException
    {
        for (RPCConnection cxn : connectionTable.values()) {
            addConnection(cxn);
        }
    }

    public LinkedHashMap<Site, Object> connect() throws RPCException {
        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("");
        }

        List<QueryThread> queryThreads = new ArrayList<QueryThread>();
        Set<Map.Entry<Site, RPCConnection>> entries = cxnTbl.entrySet();
        for (Map.Entry<Site, RPCConnection> entry : entries) {
            queryThreads.add(new QueryThread(entry.getKey(), entry.getValue(), "connect", new Object[0]));
        }

        LinkedHashMap<Site, Object> results = executeThreads(queryThreads, threadTimeoutMS, maxThreadPoolSize);

        Set<Map.Entry<Site, Object>> resultEntries = results.entrySet();

        for (Map.Entry<Site, Object> entry : resultEntries) {
            Object result = entry.getValue();
            if (result instanceof Exception) {
                RPCConnection cxn = cxnTbl.get(entry.getKey());
                cxn.setErrorMsg(((Exception) result).getMessage());
            } else {
                results.put(entry.getKey(), (cxnTbl.get(entry.getKey())).getSite());
            }
        }

        return results;
    }

    public LinkedHashMap<Site, RPCConnection> getConnections() {
        LinkedHashMap<Site, RPCConnection> result = new LinkedHashMap<Site, RPCConnection>();
        Iterator cxnIter = cxnTbl.values().iterator();
        while (cxnIter.hasNext()) {
            RPCConnection cxn = (RPCConnection) cxnIter.next();
            //if (cxn.isConnected())
            //{
            result.put(cxn.getSite(), cxn);
            //}
        }

        if (result.isEmpty()) {
            result = null;
        }
        return result;
    }

    public RPCConnection getConnection(Site site) {
        RPCConnection rtc = null;
        if (cxnTbl != null && cxnTbl.containsKey(site)) {
            rtc = cxnTbl.get(site);
        }
        return rtc;
    }

    public RPCConnection getConnection(String siteCode) {
        RPCConnection rtc = null;
        if (cxnTbl != null && !StringUtils.isEmpty(siteCode)) {
            Set<Map.Entry<Site, RPCConnection>> entries = cxnTbl.entrySet();
    
            for (Map.Entry<Site, RPCConnection> entry : entries) {
                if (entry.getKey().getSiteCode().equals(siteCode)) {
                    rtc = entry.getValue();
                    break;
                }
            }
        }
        return rtc;
    }

    public RPCConnection loginConnection() {
        return cxnTbl.get(cxnTbl.keySet().toArray()[0]);
    }

    public LinkedHashMap<Site, Object> disconnect() throws RPCException {
        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("");
        }

        if (cxnTbl.isEmpty()) {
            throw new RPCException("No connections");
        }

        // Only disconnect from the ones that are connected
        LinkedHashMap<Site, RPCConnection> myCxns = getConnections();

        List<QueryThread> queryThreads = new ArrayList<QueryThread>();
        for (Object entry : myCxns.entrySet()) {
            Site key = (Site) ((Map.Entry) entry).getKey();
            RPCConnection rpcCxn = (RPCConnection) ((Map.Entry) entry).getValue();

            queryThreads.add(new QueryThread(key, rpcCxn, "disconnect", new Object[0]));
        }

        LinkedHashMap<Site, Object> results = executeThreads(queryThreads, threadTimeoutMS, maxThreadPoolSize);


        for (Object entry : myCxns.entrySet()) {
            Site key = (Site) ((Map.Entry) entry).getKey();
            Object result = ((Map.Entry) entry).getValue();

            if (!(result instanceof Exception)) {
                results.put(key, "OK");
            }
        }

        // We only disconnected from connected connections, but we want to clear
        // the entire table
        cxnTbl = new LinkedHashMap<Site, RPCConnection>();
        return results;
    }

    protected LinkedHashMap<Site, Object> execute(String className, String methodName, Object[] args) throws RPCException {
        if (LOGGER.isDebugEnabled() && !StringUtils.isEmpty(className) && !StringUtils.isEmpty(methodName))
        {
            LOGGER.debug("className: " + className + " methodName: " + methodName);
        }

        List<QueryThread> queryThreads = new ArrayList<QueryThread>();

        Set<Map.Entry<Site, RPCConnection>> entries = cxnTbl.entrySet();
        for (Map.Entry<Site, RPCConnection> entry : entries) {
            try {
                RPCConnection cxn = entry.getValue();
                if (Class.forName("java.lang.Exception").isAssignableFrom(cxn.getClass()))
                {
                    continue;
                }
                Class daoClass = Class.forName(className, true, Thread.currentThread().getContextClassLoader());
                Class[] params = new Class[1];
                params[0] = cxn.getClass();
                Constructor cons = daoClass.getConstructor(params);
                Object dao = cons.newInstance(cxn);
                queryThreads.add(new QueryThread(entry.getKey(), dao, methodName, args));
            } catch (Exception ex) {
                throw new RPCException(ex);
            }
        }

        return executeThreads(queryThreads, threadTimeoutMS, maxThreadPoolSize);
    }

    public static LinkedHashMap<Site, Object> execute(LinkedHashMap<Site, RPCConnection> cxnTbl, String className, String methodName, Object[] args, long threadTimeout, int threadPoolSize) throws RPCException
    {
        if (LOGGER.isDebugEnabled() && !StringUtils.isEmpty(className) && !StringUtils.isEmpty(methodName))
        {
            LOGGER.debug("className: " + className + " methodName: " + methodName);
        }

        if (cxnTbl == null || cxnTbl.isEmpty()) {
            throw new RPCException("No connections!");
        }

        List<QueryThread> queryThreads = new ArrayList<QueryThread>();
        for (Object entry : cxnTbl.entrySet()) {
            try {
                Site key = (Site) ((Map.Entry) entry).getKey();

                RPCConnection cxn = (RPCConnection) ((Map.Entry) entry).getValue();
                if (Class.forName("java.lang.Exception").isAssignableFrom(cxn.getClass()))
                {
                    continue;
                }

                Class daoClass = Class.forName(className, true, Thread.currentThread().getContextClassLoader());
                Class[] params = new Class[1];
                params[0] = cxn.getClass();
                Constructor cons = daoClass.getConstructor(params);
                Object dao = cons.newInstance(cxn);
                queryThreads.add(new QueryThread(key, dao, methodName, args));
            } catch (Exception ex) {
                throw new RPCException(ex);
            }
        }

        return executeThreads(queryThreads, threadTimeout, threadPoolSize);
    }

    private static LinkedHashMap<Site, Object> executeThreads(List<QueryThread> queryThreads, long threadTimeoutMS, int maxThreadPoolSize) throws RPCException {
        if (queryThreads == null) {
            throw new IllegalArgumentException("Query threads is null");
        }

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug(ARMLog.cleanString("queryThreads (" + queryThreads.size() + "): " + queryThreads.toString()));
        }

        LinkedHashMap<Site, Object> result = new LinkedHashMap<Site, Object>();

        //if query threads is empty, return empty map
        if (!queryThreads.isEmpty()) {
            int threadPoolSize;
            if (queryThreads.size() < maxThreadPoolSize) {
                threadPoolSize = queryThreads.size();
            } else {
                threadPoolSize = maxThreadPoolSize;
            }
            
            LOGGER.debug("threadPoolSize: {}, maxThreadPoolSize: {}", threadPoolSize, maxThreadPoolSize);
    
            ExecutorService executorService = Executors.newFixedThreadPool(threadPoolSize);
    
            try {
    
                List<Future<Object>> futures = executorService.invokeAll(queryThreads, threadTimeoutMS, TimeUnit.MILLISECONDS);
    
                //futures are returned in the same sequential order as produced by the iterator for the given task list.
                Iterator<QueryThread> queryThreadItr = queryThreads.iterator();
                Iterator<Future<Object>> futuresItr = futures.iterator();
    
                while (queryThreadItr.hasNext() && futuresItr.hasNext()) {
                    QueryThread queryThread = queryThreadItr.next();
                    Future<Object> future = futuresItr.next();
    
                    try {
    
                        //if future is cancelled, assume the thread timed out
                        if (future.isCancelled()) {
                            String timeoutStr =
                                    String.format("RPC connection to VistaSite %s timed out. Connection time exceed %dms.",
                                            queryThread.getSite().getSiteCode(), threadTimeoutMS);
                            LOGGER.info(timeoutStr);
                            result.put(queryThread.getSite(), new RPCException(timeoutStr));
                        } else {
                            result.put(queryThread.getSite(), future.get());
                        }
    
                    } catch (ExecutionException e) {
                        result.put(queryThread.getSite(), e);
                    }
                }
    
            } catch (InterruptedException e) {
                throw new RPCException(e);
            } finally {
                executorService.shutdown();
            }
        }
        return result;
    }

    public String getModality() {
        return modality;
    }
}
