/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */
 
 
package gov.va.med.rpc.client;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.rpc.VistaQueryFactory;
import gov.va.med.rpc.VistaQueryFactoryUtil;
import gov.va.med.rpc.VistaResponse;
import gov.va.med.vds.vistabeans.Site;

/**
 * RPCClient is a helper class which is designed to ease the use of talking
 * RPC's with a VistA system.
 *
 * An RPC client does not need to authenticate in order to perform certain
 * queries; ie RPC "XUS INTRO MSG". Other RPCs which write data to the VistA
 * system require the RPC Client to login.
 *
 * By default, when the RPC client executes a single query using the
 * executeQuery method, the RPC client will perform the query, get the result
 * and immediately disconnect.
 *
 * To perform a query in sequence while keeping the state of the underlying
 * M stack alive, use the executeMultiQuery method. In contrast to executeQuery,
 * executeMultiQuery will keep the RPC connection open UNTIL TOLD to logout.
 *
 * Example usage can be found in the JUnit tests.
 *
 * @author vhahonuyehac
 */
public class RPCClient {
    public final static String CAPRI_CONTEXT = "DVBA CAPRI GUI";
    public final static String CPRS_CONTEXT = "OR CPRS GUI CHART";
    public final static String BCMA_CONTEXT = "PSB GUI CONTEXT - USER";
    public final static String VISTA_IMAGING_CONTEXT = "MAG WINDOWS";

    private RPCConnection rpcConnection = null;
    private String encryptedAvCode = "";
    private String brokerContext = null;
    private boolean isAuthenticated = false;
    private boolean isLoggedIn = false;
    
    private VistaQueryFactory vistaQueryFactory;

    /**
     * Instantiate an RPC client without credentials. Used for querying a
     * VistA system without providing credentials. This assumes the user knows
     * whether or not the RPC requires login in order to execute successfully.
     * @param site to connect with.
     * @throws gov.va.med.rpc.RPCException occurs when the given site code cannot be found
     */
    public RPCClient(Site site) throws RPCException {
        rpcConnection = new RPCConnection(site);

        vistaQueryFactory = VistaQueryFactoryUtil.getVistaQueryFactory(rpcConnection.getConnectionStrategy());
    }

    /**
     * Instantiates an RPC client using the given credentials. Used for reading
     * and writing data via a VistA system's RPCs.
     * @param encryptedAvCode encrypted form of the access and verify code
     * @param brokerContext the context to login as; such as
     * RPCClient.CPRS_CONTEXT
     * @param site to connect with.
     * @throws gov.va.med.rpc.RPCException occurs when the given site code cannot be found
     * or when the impl client fails to validate the given access / verify
     * code pair.
     */
    public RPCClient(String encryptedAvCode, String brokerContext, Site site) throws RPCException
    {
        super();
        if (encryptedAvCode == null || site == null || brokerContext == null) {
            throw new IllegalArgumentException("RPCClient - constructor "
                    + "arguments cannot be null.");
        }

        this.rpcConnection = new RPCConnection(site);

        connect();
        login(encryptedAvCode, brokerContext);
        disconnect();
    }

    /**
     * Instantiates a RPC client using the given credentials. Used for reading
     * and writing data via a VistA system's RPCs.
     * @param accessCode unencrypted form of the access code
     * @param verifyCode unencrypted form of the verify code
     * @param brokerContext the context to login as; such as
     * RPCClient.CPRS_CONTEXT
     * @param site VistaSite to connect with.
     * @throws gov.va.med.rpc.RPCException occurs when the impl client fails to validate the
     * given access and verify code pair or when the site code doesn't exist
     * or the broker context is invalid.
     */
    public RPCClient(String accessCode, String verifyCode, String brokerContext, Site site) throws RPCException
    {
        if (accessCode == null || verifyCode == null || site == null || brokerContext == null) {
            throw new IllegalArgumentException("RPCClient - constructor " + "arguments cannot be null.");
        }

        this.rpcConnection = new RPCConnection(site);

        connect();
        login(accessCode, verifyCode, brokerContext);
        disconnect();
    }

    public RPCClient(RPCConnection rpcConnection) {
        this.rpcConnection = rpcConnection;
    }

    private void connect() throws RPCException {
        rpcConnection.connect();
    }

    private void disconnect() throws RPCException {
        isLoggedIn = false;
        rpcConnection.disconnect();
    }

    /**
     * Login using the supplied access and verify code and broker context.
     * @param accessCode unencrypted form of the access code
     * @param verifyCode unencrypted form of the verify code
     * @param brokerContext xwb broker context, ie - RPCClient.CPRS_CONTEXT
     * @throws gov.va.med.rpc.RPCException occurs when the user is already logged in or
     * the supplied access and verify code is wrong.
     */
    public final void login(String accessCode, String verifyCode, String brokerContext) throws RPCException {
        String encryptedAvCode = vistaQueryFactory.getVistaQuery().encrypt(accessCode + ";" + verifyCode);
        login(encryptedAvCode, brokerContext);
    }

    /**
     * Login using the supplied ENCRYPTED access and verify code. By DEFAULT
     * this method will login using the CPRS context.
     * @param encryptedAvCode encrypted form of the access code + verify code
     * @throws gov.va.med.rpc.RPCException when the given access and verify code are incorrect.
     */
    public final void login(String encryptedAvCode) throws RPCException {
        login(encryptedAvCode, RPCClient.CPRS_CONTEXT);
    }

    /**
     * Login using the supplied ENCRYPTED access and verify code. The client
     * will login using the supplied broker context.
     * @param encryptedAvCode encrypted form of the access code + verify code
     * @param brokerContext the xwb context to login as after authenticating
     * @throws gov.va.med.rpc.RPCException when the given access and verify code are incorrect
     * or the broker context is invalid
     */
    public final void login(String encryptedAvCode, String brokerContext) throws RPCException {
        if (isLoggedIn()) {
            throw new RPCException("RPCClient is already logged in.");
        }

        try {
            /* Record that the user is logging into the system. */
            isLoggedIn = true;
            authenticate(encryptedAvCode, brokerContext);
        } catch (RPCException e) {
            /* Authentication failed: record that the user is NOT logged in. */
            isLoggedIn = false;
            throw new RPCException(e);
        }
    }
    
    /**
     * Called when an RPC client has completed its sequence of queries.
     * @throws gov.va.med.rpc.RPCException
     */
    public void logout() throws RPCException {
        disconnect();
    }

    public String getRPCServer() {
        return rpcConnection.getRPCEndpoint().getHost() + ":" + rpcConnection.getRPCEndpoint().getPort();
    }

    private final void authenticate(String inencryptedAvCode, String brokerContext) throws RPCException {
        String encryptedAvCode = inencryptedAvCode;
        VistaResponse vresp = null;

        VistaQuery vquery = vistaQueryFactory.getVistaQuery("XUS SIGNON SETUP");
        executeMultiQuery(vquery);

        vquery = vistaQueryFactory.getVistaQuery("XUS AV CODE");
        vquery.addParameter(VistaQuery.LITERAL, encryptedAvCode);
        vresp = executeMultiQuery(vquery);

        /* If the 7th piece is null then authentication has failed. */
        if (vresp.getPiece(StringUtils.CRLF, 7) == null) {
            encryptedAvCode = "";
            throw new RPCException(vresp.getPiece(StringUtils.CRLF, 4));
        }

        vquery = vistaQueryFactory.getVistaQuery("XWB CREATE CONTEXT");
        vquery.addEncryptedParameter(VistaQuery.LITERAL, brokerContext);
        vresp = executeMultiQuery(vquery);

        /* If the RPC doesn't return 1 then creating context has failed. */
        if (!"1".equals(vresp.getRawResponse())) {
            throw new RPCException("Failed to create XWB context: " + brokerContext);
        }

        /* If the context change is successful then inform the impl connection. */
        rpcConnection.setContext(brokerContext);
        this.brokerContext = brokerContext;

        /* The client has successfully authenticated. */
        this.isAuthenticated = true;

        /* Store the encrypted access/verify code. */
        this.encryptedAvCode = encryptedAvCode;
    }

    private VistaResponse query(VistaQuery vquery, boolean singleQuery) throws RPCException {
        if (!isConnected()) {
            connect();
        }

        if (isAuthenticated() && !isLoggedIn()) {
            login(encryptedAvCode, brokerContext);
        }

        VistaResponse vresp = rpcConnection.query(vquery);

        if (isConnected() && singleQuery) {
            disconnect();
        }

        return vresp;
    }

    /**
     * Executes a RPC query, immediately disconnects and returns the result
     * @param vistaQuery the RPC to execute
     * @return a VistaResponse object which encapsulates the RPC response
     * @throws gov.va.med.rpc.RPCException when the query fails to execute
     */
    public VistaResponse executeQuery(VistaQuery vistaQuery) throws RPCException
    {
        return query(vistaQuery, true);
    }

    /**
     * Executes a RPC query, KEEPS the RPC client connection open and returns
     * the results. This method enables a programmer to retain the M stack
     * while issuing RPCs in a procedural fashion.
     * @param vistaQuery the RPC to execute
     * @return a VistaResponse object which encapsulates the RPC response
     * @throws gov.va.med.rpc.RPCException when the query fails to execute
     */
    public VistaResponse executeMultiQuery(VistaQuery vistaQuery) throws RPCException {
        return query(vistaQuery, false);
    }

    /**
     * Returns the status of the RPC client at the socket level.
     * @return true if the impl client is still connected otherwise false.
     */
    public boolean isConnected() {
        return rpcConnection.isConnected();
    }

    /**
     * Returns whether or not the impl client has provided access and verify
     * codes which have been validated.
     * @return true if the impl client has validated the given access and verify
     * code; otherwise false.
     */
    public boolean isAuthenticated() {
        return isAuthenticated;
    }

    /**
     * Returns whether or not the current impl connection has logged in using
     * the supplied access and verify codes. Note: an impl connection does not
     * have to authenticated in order for queries to execute.
     * @return true if the impl client is currently connected and the connection
     * has logged in using the provided access and verify codes.
     */
    public final boolean isLoggedIn() {
        return isLoggedIn;
    }

    /**
     * Performs an RPC select of the patient. RPC client must be authenticated
     * before calling this method.
     * @param patientIen the vista patient ien
     * @throws gov.va.med.rpc.RPCException when the patient is not found in vista.
     */
    public void selectPatient(String patientIen) throws RPCException {
        if (!isAuthenticated()) {
            throw new IllegalStateException("RPC client requires "
                    + "authentication before calling this method.");
        }

        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWPT SELECT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);

        VistaResponse vresp = executeQuery(vquery);
        if ("-1".equalsIgnoreCase(vresp.getPiece(1))) {
            throw new RPCException("Patient not found in VistA: " + patientIen);
        }
        if (!vresp.getRawResponse().startsWith("ZZTEST")) {
            throw new RPCException("Write back is not permitted to patient: "
                    + vresp.getPiece(1));
        }
    }

}
