/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.ClinicalReminder;
import gov.va.med.vds.vistabeans.FreeTextReport;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 *
 *
 */
public class ClinicalRemindersRPC extends BaseRPC {
    private static final Logger LOGGER = LoggerFactory.getLogger(ClinicalRemindersRPC.class);
    private static final String DUE_NOW = "DUE NOW";
    private static final String NA = "NA";
    private static final String CNBD = "CNBD";
    private static final String UNKNOWN = "unknown";


    public ClinicalRemindersRPC(RPCConnection cxn) {
        super(cxn);
    }

    public ClinicalReminder[] getClinicalReminders(String patientIen) throws RPCException {
        String request = buildGetClinicalRemindersRequest(patientIen);
        String response = (String) cxn.query(request);
        return toClinicalRemindersList(response/*, patientIen*/);
    }

    public ClinicalReminder[] getClinicalReminders() throws RPCException {
        return getClinicalReminders(cxn.getPatientIen());
    }

    private String buildGetClinicalRemindersRequest(String patientIen) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORQQPX REMINDERS LIST");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        return vquery.buildMessage();
    }

    /**
     * Clinical Reminders RPC return format:
     * IEN^PRINT NAME^DUE DATE/TIME^LAST OCCURRENCE DATE/TIME
     */
    private ClinicalReminder[] toClinicalRemindersList(String response/*, String patientIen*/) {
        ClinicalReminder[] rtc;
        if ("".equals(response)) {
            rtc = new ClinicalReminder[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
    
            ArrayList<ClinicalReminder> lst = new ArrayList<ClinicalReminder>();
            ClinicalReminder rec = null;
            for (int i = 0; i < lines.length; i++) {
                if ("".equals(lines[i])) {
                    continue;
                }
                String[] flds = StringUtils.split(lines[i], StringUtils.CARET);
    
                rec = new ClinicalReminder();
                rec.setId(flds[0]);
                rec.setName(flds[1]);
                String dueDateStr = flds[2];
                String lastDoneString = flds[3];
                Calendar dueDate = null;
                Calendar lastDone = null;

                if (dueDateStr != null && !(DUE_NOW.equals(dueDateStr) || NA.equals(dueDateStr) || CNBD.equals(dueDateStr))) {
                    dueDate = Utils.getCal(Timestamp.toUtcString(dueDateStr));
                }
                if (lastDoneString != null && !UNKNOWN.equals(lastDoneString)) {
                    lastDone = Utils.getCal(Timestamp.toUtcString(lastDoneString));
                }
                rec.setDueDate(dueDate);
                rec.setDueDateString(dueDateStr);
                rec.setLastOccurrenceDate(lastDone);
                rec.setLastOccurrenceDateString(lastDoneString);
                fillCommonDataBeanValues(rec);
    
                lst.add(rec);
            }
    
            rtc = lst.toArray(new ClinicalReminder[0]);
        }
        return rtc;
    }

    public FreeTextReport getClinicalReminderDetail(String patientIen, String ien) throws RPCException {
        String request = buildGetClinicalReminderDetailRequest(patientIen, ien);
        String reportText = (String) cxn.query(request);

        FreeTextReport freeTextReport = new FreeTextReport();
        freeTextReport.setReportText(reportText);
        fillCommonDataBeanValues(freeTextReport);

        return freeTextReport;
    }

    public FreeTextReport getClinicalReminderDetail(String remindersId) throws RPCException {
        return getClinicalReminderDetail(cxn.getPatientIen(), remindersId);
    }

    private String buildGetClinicalReminderDetailRequest(String patientIen, String remindersId) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORQQPXRM REMINDER DETAIL");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        vquery.addParameter(VistaQuery.LITERAL, remindersId);
        return vquery.buildMessage();
    }


    /**
     * Uses the VPR RPC with "reminders" as the type.
     *
     * @return The list of appointments for the specified date range
     * @throws RPCException
     */
    public ClinicalReminder[] getClinicalRemindersVPR() throws RPCException {
        return getClinicalRemindersVPR(cxn.getPatientIen());
    }

    /**
     * Uses the VPR RPC with "reminders" as the type.
     *
     * @param patientIen The patient IEN
     * @return The list of appointments for the specified date range
     * @throws RPCException
     */
    public ClinicalReminder[] getClinicalRemindersVPR(String patientIen) throws RPCException {
        String request = buildGetRemindersRequestVPR(patientIen);
        String response = (String) cxn.query(request);

        ClinicalReminder[] reminders = parseReminders(response/*, false*/);

        return reminders;
    }

    /**
     * Builds the VPR RPC request string using "appointments" as the type. The problems contain the list of
     * appointments. The RPC will fail if the "VPR APPLICATION PROXY" context is not set. That is set when
     * setting up the connection.
     *
     * @param patientIen Patient IEN
     * @return The RPC response message string
     */
    private String buildGetRemindersRequestVPR(String patientIen) {
        return VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, VPRUtil.VPRType.REMINDERS);
    }

    private ClinicalReminder[] parseReminders(String response/*, boolean admissionsOnly*/) throws RPCException {
        ClinicalReminder[] rtc;
        if ("".equals(response)) {
            rtc = new ClinicalReminder[0];
        } else {
    
            StringBuilder sbld = new StringBuilder(response);
            sbld.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    
            ArrayList<ClinicalReminder> remindersList = new ArrayList<ClinicalReminder>();
    
            final String XPATH_REMINDER = "//reminder";
    
            Document document = null;
            XPathExpression exprReminder = null;
    
            try {
                document = parseXMLDocument(sbld.toString());
                XPathFactory xPathFactory = XPathFactory.newInstance();
                XPath xpath = xPathFactory.newXPath();
                exprReminder = xpath.compile(XPATH_REMINDER);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            NodeList remindersNodeList = null;
            try {
                remindersNodeList = (NodeList) exprReminder.evaluate(document, XPathConstants.NODESET);
            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }
    
            List<VPRRemindersParser> remindersParserList = new ArrayList<VPRRemindersParser>();
            int reminderNodeListLength = remindersNodeList.getLength();
            for (int reminderNum = 0; reminderNum < reminderNodeListLength; reminderNum++) {
                Node reminderNode = remindersNodeList.item(reminderNum);
                reminderNode.getParentNode().removeChild(reminderNode);
    
                remindersParserList.add(new VPRRemindersParser(reminderNode));
            }
    
            ExecutorService executorService = Executors.newFixedThreadPool(5);
            List<Future<ClinicalReminder>> futures = null;
            try {
                futures = executorService.invokeAll(remindersParserList, 2, TimeUnit.MINUTES);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse Reminders results : reminder node parser futures list is null.");
                }
            } catch (InterruptedException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
                throw new RPCException(e);
            } finally {
                executorService.shutdown();
            }
    
            Iterator<Future<ClinicalReminder>> futureIterator = futures.iterator();
    
            while (futureIterator.hasNext()) {
                try {
                    Future<ClinicalReminder> future = futureIterator.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR ClinicalReminder Data");
                    } else {
                        ClinicalReminder reminder = future.get();
    
                        // exclude null results
                        if (reminder != null) {
                            remindersList.add(reminder);
                        }
                    }
                } catch (InterruptedException ie) {
                    RPCUtil.logErrorMessage(LOGGER, ie);
                } catch (ExecutionException ee) {
                    RPCUtil.logErrorMessage(LOGGER, ee);
                }
            }
    
            rtc = remindersList.toArray(new ClinicalReminder[0]);
        }
        return rtc;
    }

    private class VPRRemindersParser implements Callable<ClinicalReminder> {
        private Node reminderNode;

        public VPRRemindersParser(Node reminderNode) {
            this.reminderNode = reminderNode;
        }

        public ClinicalReminder call() throws Exception {

            final String XPATH_CLASS = "./class/@name";
            final String XPATH_DETAIL = "./detail";
            final String XPATH_DUE = "./due/@value";
            final String XPATH_ID = "./id/@value";
            final String XPATH_LAST_DONE = "./lastDone/@value";
            final String XPATH_NAME = "./name/@value";
            final String XPATH_STATUS = "./status/@value";
            final String XPATH_SUMMARY = "./summary";

            XPathExpression exprClass = null;
            XPathExpression exprDetail = null;
            XPathExpression exprDue = null;
            XPathExpression exprId = null;
            XPathExpression exprLastDone = null;
            XPathExpression exprName = null;
            XPathExpression exprStatus = null;
            XPathExpression exprSummary = null;

            XPathFactory xPathFactory = XPathFactory.newInstance();
            XPath xpath = xPathFactory.newXPath();

            try {
                exprClass = xpath.compile(XPATH_CLASS);
                exprDetail = xpath.compile(XPATH_DETAIL);
                exprDue = xpath.compile(XPATH_DUE);
                exprId = xpath.compile(XPATH_ID);
                exprLastDone = xpath.compile(XPATH_LAST_DONE);
                exprName = xpath.compile(XPATH_NAME);
                exprStatus = xpath.compile(XPATH_STATUS);
                exprSummary = xpath.compile(XPATH_SUMMARY);
            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }

            String classString = RPCUtil.evaluateNodeToString(reminderNode, exprClass, LOGGER);
            String detail = RPCUtil.evaluateNodeToString(reminderNode, exprDetail, LOGGER);
            String dueString = RPCUtil.evaluateNodeToString(reminderNode, exprDue, LOGGER);
            String id = RPCUtil.evaluateNodeToString(reminderNode, exprId, LOGGER);
            String lastDoneString = RPCUtil.evaluateNodeToString(reminderNode, exprLastDone, LOGGER);
            String name = RPCUtil.evaluateNodeToString(reminderNode, exprName, LOGGER);
            String status = RPCUtil.evaluateNodeToString(reminderNode, exprStatus, LOGGER);
            String summary = RPCUtil.evaluateNodeToString(reminderNode, exprSummary, LOGGER);


            ClinicalReminder reminder = new ClinicalReminder();
            fillCommonDataBeanValues(reminder);

            Calendar due = null;
            if (dueString != null && !(DUE_NOW.equals(dueString) || NA.equals(dueString) || CNBD.equals(dueString))) {
                due = Timestamp.toDateTime(dueString);
            }
            Calendar lastDone = null;
            if (lastDoneString != null && !UNKNOWN.equals(lastDoneString)) {
                lastDone = Timestamp.toDateTime(lastDoneString);
            }
            reminder.setLastOccurrenceDate(lastDone);
            reminder.setLastOccurrenceDateString(lastDoneString);
            reminder.setDueDate(due);
            reminder.setDueDateString(dueString);

            reminder.setReminderClass(classString);
            reminder.setDetail(detail);
            reminder.setId(id);
            reminder.setName(name);
            reminder.setStatus(status);
            reminder.setDescription(summary);

            return reminder;
        }
    }

}
