/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.patient.dao.AllergyDao;
import gov.va.med.vds.vistabeans.Allergy;
import gov.va.med.vds.vistabeans.AllergyDetail;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class AllergyDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(AllergyDataService.class.getName());

    /**
     * Retrieves list of patient allergies.
     * @param queryBean Query parameters.
     *                  Required params: User, Patient
     * @return A list of patient allergies.
     * @throws VDSException if there was an error..
     */
    public List<Allergy> getPatientAllergies(QueryBean queryBean) throws VDSException {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupVPRContextMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> allergyMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_ALLERGY,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_ALLERGIES, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = allergyMap.entrySet();
            List<Allergy> allergyList = new ArrayList<Allergy>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object allergies = entry.getValue();
                if (allergies instanceof Exception) {
                    allergies = generateConnectionUnavailableBean(entry.getKey(), (Exception) allergies);
                }
                allergyList.addAll(Arrays.asList((Allergy[]) allergies));
            }

            return allergyList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    /**
     * Retrieves patient allergy details from the provided location.
     * @param queryBean Query parameters.
     *                  Required params: User, Patient, itemId, recordSiteCode
     * @return AllergyDetail containing allergy report text.
     * @throws VDSException if there was an error..
     */
    public AllergyDetail getAllergyDetail(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        RPCConnection cxn = null;

        try {
            String itemId = queryBean.getItemId();
            String siteCode = queryBean.getRecordSite().getSiteCode();
            if (itemId == null || itemId.length() <= 0) {
                throw new IllegalArgumentException("Item Id cannot be null or empty");
            } else if (StringUtils.isEmpty(siteCode)) {
                throw new IllegalArgumentException("Missing record site code");
            }
            cxn = setupSingleRemoteSourceQuery(queryBean);
            AllergyDao allergyDao = new AllergyDao(cxn);
            return allergyDao.getAllergyDetail(queryBean.getItemId());
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxn != null) {
                try {
                    cxn.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private Allergy[] generateConnectionUnavailableBean(Site site, Exception except) {
        
        Allergy allergy = new Allergy();
        allergy.setAllergyName(CONNECTION_UNAVAILABLE);
        allergy.setSite(site);
        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            allergy.setAllergyName(UNABLE_TO_RETRIEVE);
        } else {
            LOGGER.error(except.getMessage(), except);        //log if any other exception type
        }
        return new Allergy[]{allergy};
    }
}
