/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.patient.dao.ProblemsDao;
import gov.va.med.vds.vistabeans.Problem;
import gov.va.med.vds.vistabeans.ProblemDetail;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class ProblemsDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(ProblemsDataService.class.getName());

    public List<Problem> getPatientProblemList(QueryBean queryBean) throws VDSException {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            if (StringUtils.isEmpty(queryBean.getStatus())) {
                throw new IllegalArgumentException("Problem type is null or empty");
            }
            msq = setupVPRContextMultiSourceQuery(queryBean); //, VistaConstants.CPRS_CONTEXT);
            LinkedHashMap<Site, Object> problemsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_PROBLEMS,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_PROBLEM_LIST, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = problemsMap.entrySet();
            List<Problem> problemsList = new ArrayList<Problem>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object problems = entry.getValue();
                if (problems instanceof Exception) {
                    problems = generateConnectionUnavailableBean(entry.getKey(), (Exception) problems);
                }
                problemsList.addAll(Arrays.asList((Problem[]) problems));
            }

            return problemsList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public ProblemDetail getProblemDetail(QueryBean queryBean) throws VDSException {
        validateQueryBean(queryBean);

        String problemId = queryBean.getItemId();
        String recordSiteCode = queryBean.getRecordSite().getSiteCode();

        RPCConnection cxn = null;

        try {
            if (problemId == null || problemId.length() <= 0) {
                throw new IllegalArgumentException("Item Id cannot be null or empty");
            } else if (StringUtils.isEmpty(recordSiteCode)) {
                throw new IllegalArgumentException("Missing record site code");
            }

            cxn = setupSingleRemoteSourceQuery(queryBean);
            ProblemsDao problemDao = new ProblemsDao(cxn);

            ProblemDetail detail = problemDao.getProblemDetail(queryBean);

            return detail;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxn != null) {
                try {
                    cxn.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private Problem[] generateConnectionUnavailableBean(Site site, Exception except) {
        Problem problem = new Problem();
        problem.setDescription(CONNECTION_UNAVAILABLE);
        problem.setLastModifiedDate(Utils.backDate().toGregorianCalendar());
        problem.setSite(site);

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            problem.setDescription(UNABLE_TO_RETRIEVE);
        } else {
            LOGGER.error(except.getMessage(), except);        //log if any other exception type
        }
        return new Problem[] {problem};
    }
}
