/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */

package gov.va.med.vds.provider.dao.cache.model;

import gov.va.med.vds.common.AppConfig;
import gov.va.med.vds.vistabeans.DataBean;
import gov.va.med.vds.vistabeans.PatientDemographics;
import gov.va.med.vds.vistabeans.Site;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ModelData {
    public static final int ERRCODE_Exception = -999;
    public static final int ERRCODE_Error = -1;
    public static final int ERRCODE_Success = 0;
    public static final int ERRCODE_More = 1;
    
    protected static final AppConfig APP_CONFIG = AppConfig.getInstance();
    protected static final String AGENCY_VA = APP_CONFIG.getAgencyVA();
    protected static final String PROTOCOL_CACHE = APP_CONFIG.getProtocolCache();
    
    private static final Logger LOGGER = LoggerFactory.getLogger(ModelData.class);

    // All COMMON error code values for the attribute "mErrorCode".
    // Error code values specific to subclasses that extend this class should
    // define their own values using the syntax "ERRCODE_Nnn" (something
    // descriptive).

    // This value is accessable by the JSPs.
    // It contains the error code set by the "getData()" function of
    // the subclass which extends this class.
    // >= 0 Success
    // < 0 Failure
    protected int mErrorCode = 0;

    // This value is accessable by the JSPs. If it is NOT null,
    // it contains an error message set by the "getData()" function of
    // the subclass which extends this class. "mErrorCode" should also be
    // set to < 0 and appended to this string in the case of an error.
    protected String mErrorText = null;

    // When model-side methods on "session" are successful;
    // - mTable is populated
    // - getData() sets mErrorText = null;
    // - getData() sets mErrorCode >= 0
    //
    // When model-side methods on "session" are NOT successful;
    // - mTable is null
    // - getData() sets mErrorText = value retrieved from variant;
    // - getData() sets mErrorCode < 0
    protected ResultSet mTable = null;

    protected List mObjectList = null;
    // Variant data which can potentially be returned by model-side calls
    // (has a public accessor). "setErrorTextAndStatusCode()" sets this.
    protected String mVariantData = null;

    protected Site mSite = null;

    // Contructors of classes that subclass this must call "super()"
    protected ModelData(Site site, ResultSet table) {
        mTable = table;
        mSite = site;
    }

    // Contructors of classes that subclass this must call "super()"
    protected ModelData(Site site, List objList) {
        mObjectList = objList;
        mSite = site;
    }

    protected void fillCommonDataBeanValues(DataBean bean) {
        bean.setSite(mSite);
        bean.setSourceProtocol(PROTOCOL_CACHE);
    }

    protected void fillCommonDataBeanValues(PatientDemographics patDemoBean) {
        patDemoBean.setSite(mSite);
        patDemoBean.setSourceProtocol(PROTOCOL_CACHE);
    }

    /**
     * Retrieve data from the table...
     */
    protected String getCell(int row, int cell) {
        String rtc = null;
        mErrorText = null;
        if (mTable != null) {
            int col = cell + 1;
            try {
                rtc = mTable.getString(col);
            } catch (SQLException e) {
                mErrorText = e.getMessage();
                LOGGER.error(e.getMessage());
            }
        }
        return rtc;
    }

    /**
     * Retrieve data from the table...
     */
    protected String getCell(ResultSet rset, int row, int cell) {
        String rtc = null;
        mErrorText = null;
        if (rset != null) {
            int col = cell + 1;
            try {
                rtc = rset.getString(col);
            } catch (SQLException e) {
                mErrorText = e.getMessage();
                LOGGER.error(e.getMessage());
            }
        }
        return rtc;
    }

    /**
     * Called by "xxxAccess" objects - Sets the "mErrorText" and "mErrorCode"
     * VALUES BASED ON "variantStatusCode". - Logs the messages for errors.
     */
    public void setErrorTextAndStatusCode(String variantValue,
            int variantStatusCode) {
        mErrorText = null;
        if (variantStatusCode == 0) { // RULE A; 0 Indicates success.
            mVariantData = variantValue;
            mErrorCode = ERRCODE_Success;
        } else if (variantStatusCode == 1) { // RULE A; 1 Indicates success,
                                                // more data.
            mVariantData = variantValue;
            mErrorCode = ERRCODE_More;
        } else if (variantStatusCode > 1) { // RULE A; Other codes > 1 indicate
                                            // success, plus other status.
            mVariantData = variantValue;
            mErrorCode = ERRCODE_Success;
        } else { // RULE B; Other codes < 0 indicate error.
            mVariantData = null;
            mErrorText = variantValue;
            mErrorCode = ERRCODE_Error;

            mTable = null;

            logErrorText(variantStatusCode);
        }
    }

    /**
     * Append "mErrorCode" and "status" to "mErrorText".
     */
    private void logErrorText(int status) {
        if (mErrorText != null) {
            // Set final error text...
            mErrorText = mErrorText + " (mErrorCode = " + mErrorCode
                    + ", status = " + status + ")";

            // Log the error... Coming soon.
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug(mErrorText);
            }
        }
    }

    // ----------------- PUBLIC METHODS BELOW -------------------

    public int getErrorCode() {
        return mErrorCode;
    }

    /**
     * If "mErrorText" is NOT null, an error occured during "getData()" and thus
     * mTable should be null and the following "get" functions should return
     * null.
     */
    public String getErrorText() {
        return mErrorText;
    }

    /**
     *
     */
//    public java.util.Iterator createIterator() {
//        return new DataModelIterator(this);
//    }

    public int getColumns() {
        int count = 0;

        try {
            ResultSetMetaData rsmd = mTable.getMetaData();
            count = rsmd.getColumnCount();
        } catch (SQLException e) {
            mErrorText = e.getMessage();
            LOGGER.error(e.getMessage());
        }
        return count;
    }
}