/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
package gov.va.med.vds.webservice;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.concurrent.ConcurrentHashMap;

import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * VistA Data Status.
 *
 * @author <a href="mailto:kweller@hawaiirg.com">kweller</a>
 * @version 1.0
 *
 */
public class VistaDataStatus extends HttpServlet {
    /**
     * Class UID.
     */
    private static final long serialVersionUID = 5494436108721278388L;

    /**
     * Startup time.
     */
    private static Date startDate = new Date();
    
    /**
     * Number of calls.
     */
    private static ConcurrentHashMap<String, Long> callCount = new ConcurrentHashMap<String, Long>();
    
    /**
     * Number of errors.
     */
    private static ConcurrentHashMap<String, Long> callErrorCount = new ConcurrentHashMap<String, Long>();
    
    /**
     * Error timestamp.
     */
    private static ConcurrentHashMap<String, Date> callErrorTimeStamp = new ConcurrentHashMap<String, Date>();
    
    /**
     * Time of last call.
     */
    private static ConcurrentHashMap<String, Date> callTimeStamp = new ConcurrentHashMap<String, Date>();
    
    /**
     * Time for calls to complete.
     */
    private static ConcurrentHashMap<String, Long> callTotalTime = new ConcurrentHashMap<String, Long>();
    
    /**
     * Minimum time for calls to complete.
     */
    private static ConcurrentHashMap<String, Long> callMinTime = new ConcurrentHashMap<String, Long>();
    
    /**
     * Maximum time for calls to complete.
     */
    private static ConcurrentHashMap<String, Long> callMaxTime = new ConcurrentHashMap<String, Long>();
    
    /**
     * Called when a get request needs to be processed. Used to check status.
     *
     * @param request request object
     * @param response response object
     * 
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response) {
        String rtc;
        try {
            rtc = getHTMLStatus();
            response.setContentType("text/html;charset=UTF-8");
            response.setCharacterEncoding("UTF-8");
            response.getWriter().write("<html><head><title>VistaDataService Status</title></head><body><H1>VistaDataService</H1><hr><br />" + rtc + "<br /></body></html>");
        } catch (Throwable t) {
            try {
                response.getWriter().write("<html><head><title>ERROR</title></head><body><hr><br />ERROR<br /></body></html>");
            } catch (Throwable t2) {
                rtc = null;
            }
        }
    }
    
    /**
     * Record the statistics of a call.
     * 
     * @param method method called
     * @param cache did return come from cache (true or false)
     * @param error error flag
     * @param millis number of milliseconds for call to complete
     */
    public static synchronized void updateStatistics(String method, boolean error, long millis) {
        Date current = new Date();
        Long cnt = callCount.get(method);
        if (cnt == null) {
            cnt = Long.valueOf(0);
        }
        cnt++;
        callCount.put(method, cnt);
        Long min = callMinTime.get(method);
        if (min == null) {
            min = Long.valueOf(99999999);
        }
        if (millis < min && millis > 0) {
            callMinTime.put(method, Long.valueOf(millis));
        }
        Long max = callMaxTime.get(method);
        if (max == null) {
            max = Long.valueOf(0);
        }
        if (millis > max && millis > 0) {
            callMaxTime.put(method, Long.valueOf(millis));
        }
        Long tcnt = callTotalTime.get(method);
        if (tcnt == null) {
            tcnt = Long.valueOf(0);
        }
        tcnt += millis;
        callTotalTime.put(method, tcnt);
        if (error) {
            Long ecnt = callErrorCount.get(method);
            if (ecnt == null) {
                ecnt = Long.valueOf(0);
            }
            ecnt++;
            callErrorCount.put(method, ecnt);
            callErrorTimeStamp.put(method, current);
        }
        callTimeStamp.put(method, current);
    }
    
    /**
     * Clear Statistics.
     */
    public static void resetStatus() {
        callCount.clear();
        callMinTime.clear();
        callMaxTime.clear();
        callErrorCount.clear();
        callErrorTimeStamp.clear();
        callTimeStamp.clear();
        callTotalTime.clear();
    }
    
    /**
     * Get service status.
     * 
     * @return status
     */
    public static String getStatus() {
        StringBuffer rtc = new StringBuffer(1024);
        rtc.append("VistaDataService - Running Since: ");
        rtc.append(startDate.toString());
        rtc.append("\n\nCall Type                Count     Errors    Min(ms)   Max(ms)   Avg(ms)   Date Last Call               Date Last Error\n------------------------------------------------------------------------------------------------------------------------------------\n");
        if (callCount == null || callCount.isEmpty()) {
            rtc.append("No calls received.");
        } else {
            Object[] key = callCount.keySet().toArray();
            Arrays.sort(key);
            for (int i = 0;  i < key.length; ++i) {
                StringBuffer row = new StringBuffer();
                String name = (String) key[i];
                Long count = callCount.get(name);
                Long error = callErrorCount.get(name);
                if (error == null) {
                    error = Long.valueOf(0);
                }
                Long min = callMinTime.get(name);
                Long max = callMaxTime.get(name);
                
                Date lastCall = callTimeStamp.get(name);
                Date lastErrorCall = callErrorTimeStamp.get(name);
                Long totalTime = callTotalTime.get(name);
                if (totalTime == null) {
                    totalTime = Long.valueOf(0);
                }
                if (count == null || count.longValue() == 0L) {
                    row.append(name);
                    row.append(": N/A\n");
                } else {
                    row.append(name);
                    row.append(' ');
                    while (row.length() < 25) {
                        row.append(' ');
                    }
                    row.append(count.toString());
                    row.append(' ');
                    while (row.length() < 35) {
                        row.append(' ');
                    }
                    if (error != null) {
                        row.append(error.toString());                        
                    }
                    row.append(' ');
                    while (row.length() < 45) {
                        row.append(' ');
                    }
                    if (min != null) {
                        row.append(min.toString());                        
                    }
                    row.append(' ');
                    while (row.length() < 55) {
                        row.append(' ');
                    }
                    if (max != null) {
                        row.append(max.toString());                        
                    }
                    row.append(' ');
                    while (row.length() < 65) {
                        row.append(' ');
                    }
                    if (count.longValue() <= 0) {
                        row.append('0');
                    } else {
                        int avg = (int) (totalTime.longValue() / (count.longValue()));
                        row.append(avg);
                    }
                    row.append(' ');
                    while (row.length() < 75) {
                        row.append(' ');
                    }
                    if (lastCall != null) {
                        row.append(lastCall.toString());
                    }
                    if (lastErrorCall != null) {
                        row.append(' ');
                        row.append(lastErrorCall.toString());                        
                    }
                    rtc.append(row.toString());
                    rtc.append('\n');
                }
            }
        }
        return rtc.toString();
    }

    /**
     * Get service status.
     * 
     * @return status
     */
    public static String getHTMLStatus() {
        StringBuffer rtc = new StringBuffer(1024);
        rtc.append("VistaDataService - Running Since: ");
        rtc.append(startDate.toString());
        rtc.append("<br />Running on: ");
        rtc.append(getComputerId());
        rtc.append("<br /><br />");
        if (callCount == null || callCount.isEmpty()) {
            rtc.append("<br />No calls received.<br />");
        } else {
            rtc.append("<table><thead><tr style=\"text-align:left\"><th>Call Type</th><th>Count&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</th><th>Errors&nbsp;&nbsp;&nbsp;&nbsp;</th><th>Min(ms)&nbsp;&nbsp;&nbsp;</th><th>Max(ms)&nbsp;&nbsp;&nbsp;</th><th>Avg(ms)&nbsp;&nbsp;&nbsp;</th><th>Date Last Call</th><th>Date Last Error</th></tr></thead><tbody>");
            Object[] key = callCount.keySet().toArray();
            Arrays.sort(key);
            for (int i = 0; i < key.length; ++i) {
                String name = (String) key[i];
                Long count = callCount.get(name);
                Long error = callErrorCount.get(name);
                if (error == null) {
                    error = Long.valueOf(0);
                }
                Long min = callMinTime.get(name);
                Long max = callMaxTime.get(name);
                Date lastCall = callTimeStamp.get(name);
                Date lastError = callErrorTimeStamp.get(name);
                Long totalTime = callTotalTime.get(name);
                if (totalTime == null) {
                    totalTime = Long.valueOf(0);
                }
                if (count == null || count.longValue() == 0L) {
                    continue;
                }
                rtc.append("<tr><td>");
                rtc.append(name);
                rtc.append("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td><td>");
                rtc.append(count.toString());
                rtc.append("</td>");

                if (error != null) {
                    rtc.append("<td>");
                    rtc.append(error.toString());
                    rtc.append("</td>");
                } else {
                    rtc.append("<td></td>");
                }
                if (min != null) {
                    rtc.append("<td>");
                    rtc.append(min.toString());
                    rtc.append("</td>");
                } else {
                    rtc.append("<td></td>");
                }
                if (max != null) {
                    rtc.append("<td>");
                    rtc.append(max.toString());
                    rtc.append("</td>");
                } else {
                    rtc.append("<td></td>");
                }
                if (count.longValue() <= 0) {
                    rtc.append("<td>0</td>");
                } else {
                    int avg = (int) (totalTime.longValue() / (count.longValue()));
                    rtc.append("<td>");
                    rtc.append(avg);
                    rtc.append("</td>");
                }
                if (lastCall != null) {
                    rtc.append("<td>");
                    rtc.append(lastCall.toString());
                    rtc.append("</td>");
                } else {
                    rtc.append("<td></td>");
                }
                if (lastError != null) {
                    rtc.append("<td>");
                    rtc.append(lastError.toString());
                    rtc.append("</td>");
                } else {
                    rtc.append("<td></td>");
                }
                rtc.append("</tr>\n");
            }
            rtc.append("</tbody>\n</table>\n");
        }
        return rtc.toString();
    }

    /**
     * Get computer id.
     * 
     * @return id
     */
    private static String getComputerId() {
        StringBuffer idstring = new StringBuffer();

        String hostname = System.getProperty("COMPUTERNAME", System.getenv().get("COMPUTERNAME"));
        if (hostname == null || hostname.length() == 0) {
            hostname = System.getProperty("HOSTNAME", System.getenv().get("HOSTNAME"));
        }
        if (hostname != null && hostname.length() > 0) {
            idstring.append(hostname);
            idstring.append('|');
        }

        try {
            Enumeration<NetworkInterface> nets = NetworkInterface.getNetworkInterfaces();
            for (NetworkInterface netint : Collections.list(nets)) {
                if (netint.isUp() && (!netint.isLoopback()) && (!netint.isPointToPoint()) && (!netint.isVirtual())) {
                    Enumeration<InetAddress> inetAddresses = netint.getInetAddresses();
                    for (InetAddress inetAddress : Collections.list(inetAddresses)) {
                        if (inetAddress instanceof Inet4Address) {
                            if (idstring.length() == 0) {
                                idstring.append(inetAddress.getHostName());
                                idstring.append('|');
                            }
                            idstring.append(inetAddress.getHostAddress());
                            idstring.append('|');
                        }
                    }
//                    byte[] mac = netint.getHardwareAddress(); 
//                    if (mac != null) { 
//                        for (int i = 0; i < mac.length; i++) { 
//                            if (i > 0) {
//                                idstring.append('-');    
//                            }
//                            idstring.append(String.format("%02X", mac[i]));
//                        } 
//                    }
//                    idstring.append('|');
                }
            }
        } catch (SocketException ex) {
            idstring.append('|');
        }
        return idstring.toString();
    }
    
//    public static void main(String[] args) {
//        VistaDataStatus.updateStatistics("TEST1", true, 100);
//        VistaDataStatus.updateStatistics("TEST1", true, 100);
//        VistaDataStatus.updateStatistics("TEST1", false, 100);
//        VistaDataStatus.updateStatistics("TEST1", false, 100);
//        VistaDataStatus.updateStatistics("TEST1", false, 100);
//
//        VistaDataStatus.updateStatistics("TEST2", true, 200);
//        VistaDataStatus.updateStatistics("TEST2", true, 200);
//        VistaDataStatus.updateStatistics("TEST2", false, 200);
//        VistaDataStatus.updateStatistics("TEST2", false, 200);
//
//        VistaDataStatus.updateStatistics("TEST2", false, 200);
//        VistaDataStatus.updateStatistics("TEST3", false, 300);
//
//        System.out.println(VistaDataStatus.getStatus());
//    }
}
