/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartAdmission;
import gov.va.med.jmeadows.smartbeans.patient.SmartDiagnosis;
import gov.va.med.jmeadows.smartbeans.patient.SmartProcedure;
import gov.va.med.vds.webservice.AdmissionDetail;
import gov.va.med.vds.webservice.Diagnosis;
import gov.va.med.vds.webservice.PatientAdmission;
import gov.va.med.vds.webservice.Procedure;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Patient admission data access object
 */
public class AdmissionDao extends DaoBase {

    /**
     * Retrieves a list of patient admission records from all available patient locations.
     *
     * @param queryBean Query parameters
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return A list of patient admission records.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientAdmission> getPatientAdmissions(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatAdmissions", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                MultiLocationQueryConfig patAdmissionsConfig = new PatientMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_bhieRelayServiceHandler)
                        .dataServiceMethod("getPatientAdmissions")
                        .dataServiceMethodArgs(bhieQueryBean)
                        .smartBeanMappingMethod("mapAdmissionsFromBHIE")
                        .location(bhieLoc)
                        .build();

                configs.add(patAdmissionsConfig);

            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getPatientAdmissions")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        Object[] admsObjs = MultiLocationQuery.execute(
                configs, new SmartAdmission(), queryBean);

        List<PatientAdmission> patientAdmissions = null;

        if (admsObjs != null) {
            patientAdmissions = Arrays.asList((PatientAdmission[]) admsObjs);
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, patientAdmissions, CodingSystem.SNOMED);

        }

        return patientAdmissions;
    }


    /**
     * Retrieves patient admission record details from the provided record location.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     * recordSiteCode
     * itemId
     *
     * @return AdmissionDetail bean.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public AdmissionDetail getAdmissionDetails(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.PATIENT,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        auditQuery("AdmissionDetail", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        AdmissionDetail admissionDetail = null;

        Site recordSite = getSite(queryBean.getRecordSiteCode());

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())
                && LocationHelper.vistaLocationsContainSiteCode(recordSite, queryBean.getPatient().getVistaSites())) {

                admissionDetail = sm_vistaHandler.getAdmissionDetails(VistaBeanFactory.create(queryBean));
                break;
            }
        }

        return admissionDetail;
    }

    /**
     * Retrieves a list of admission diagnosis from all available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return A list of admission diagnosis records.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Diagnosis> getAdmissionDiagnosis(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("AdmissionDiagnosis", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getAdmissionDiagnosis")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        Object[] diagObjs = MultiLocationQuery.execute(
                configs, new SmartDiagnosis(), queryBean);

        List<Diagnosis> diagnosisList = null;

        if (diagObjs != null) {
            diagnosisList = Arrays.asList((Diagnosis[]) diagObjs);
        }

        return diagnosisList;
    }

    /**
     * Retrieves a list of admission procedures from all available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return A list of admission procedures.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Procedure> getAdmissionProcedures(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("AdmissionProcedures", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getAdmissionProcedures")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        Object[] procObjs = MultiLocationQuery.execute(
                configs, new SmartProcedure(), queryBean);

        List<Procedure> procedureList = null;

        if (procObjs != null) {
            procedureList = Arrays.asList((Procedure[]) procObjs);
        }

        return procedureList;
    }
}
