/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartAppointment;
import gov.va.med.jmeadows.smartbeans.patient.SmartEncounter;
import gov.va.med.vds.webservice.Encounter;
import gov.va.med.vds.webservice.FreeTextReport;
import gov.va.med.vds.webservice.PatientAppointments;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Patient appointments data access object
 */
public class AppointmentsDao extends DaoBase {

    /**
     * Retrieves a list of patient appointments from all available patient locations.
     *
     * @param queryBean Query parameters
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T + 1 year).
     *
     * @return A list of patient appointments.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientAppointments> getPatientAppointments(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatAppointments", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(Utils.getDate(Utils.getRelativeDate(1, 0, 0)))));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includdedSources = new ArrayList<String>();
                includdedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includdedSources, null);

                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_bhieRelayServiceHandler)
                            .dataServiceMethod("getPatientAppointments")
                            .dataServiceMethodArgs(bhieQueryBean)
                            .smartBeanMappingMethod("mapAppointmentsFromBHIE")
                            .location(bhieLoc)
                            .build()
                );
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getPatientAppointments")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        Object[] aptObjs = MultiLocationQuery.execute(
                configs, new SmartAppointment(), queryBean);

        List<PatientAppointments> appointments = null;

        if (aptObjs != null) {
            appointments = Arrays.asList((PatientAppointments[]) aptObjs);
        }

        return appointments;
    }

    /**
     * Retrieves a list of patient encounters .
     *
     * @param queryBean Query parameters
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return A list of patient encounters from AHLTA.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Encounter> getPatientEncounters(JMeadowsQuery queryBean) throws JMeadowsException
    {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatEncounters", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_bhieRelayServiceHandler)
                            .dataServiceMethod("getPatientEncounters")
                            .dataServiceMethodArgs(bhieQueryBean)
                            .smartBeanMappingMethod("mapEncountersFromBHIE")
                            .location(bhieLoc)
                            .build()
                );
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientEncounters")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
            }
        }

        Object[] encounterObjs = MultiLocationQuery.execute(
                configs, new SmartEncounter(), queryBean);

        List<Encounter> encounters = null;

        if (encounterObjs != null) {
            encounters = Arrays.asList((Encounter[]) encounterObjs);

            //map problem concept codes
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, encounters, CodingSystem.SNOMED);

        }

        return encounters;
    }


    /**
     * Retrieves a Encounters Report from the specified site.
     *
     * @param queryBean Query parameters
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     * recordSiteCode
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T + 1 year).
     *
     * @return A report of patient encounters.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */

    public FreeTextReport getPatientEncountersReport(JMeadowsQuery queryBean) throws JMeadowsException {


        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.PATIENT,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.DATE_RANGE);

        auditQuery("EncountersReport", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        FreeTextReport encountersReport = null;

        Site recordSite = getSite(queryBean.getRecordSiteCode());

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())
                    && LocationHelper.vistaLocationsContainSiteCode(recordSite, queryBean.getPatient().getVistaSites())) {

                encountersReport = sm_vistaHandler.getEncountersReport(VistaBeanFactory.create(queryBean));
                break;
            }

        }

        return encountersReport;
    }
}
