/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.vhahon.common.Utils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.SQLException;


/**
 * Audit data access object
 */
public class AuditDao extends DaoBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(AuditDao.class);

    /**
     * Sends audit log to audit sql database.
     *
     * @param siteId Site identifier
     * @param userIen user (user) identifier
     * @param userNPI user (user) national user identifier
     * @param userName user (user) name
     * @param PatId  patient ien.
     * @param Category audit category
     * @param requestingApp requesting application
     * @param StartDate  query start date (defaults to T)
     * @param EndDate    query end date   (defaults to T)
     * @throws JMeadowsException if an error occurs
     */
    public void setAudit(String siteId, String userIen, String userNPI,
                         String userName, String PatId, String Category,
                         String requestingApp, String StartDate, String EndDate) throws JMeadowsException {
        Connection sqlConn = null;
        CallableStatement cstmt = null;

        java.sql.Date stDate = null;
        java.sql.Date enDate = null;

        if (Utils.getDate(StartDate) != null) {
            stDate = new java.sql.Date(Utils.getDate(StartDate).getTime());
        }
        if (Utils.getDate(EndDate) != null) {
            enDate = new java.sql.Date(Utils.getDate(EndDate).getTime());
        }

        try {
            sqlConn = sm_dbConnectionPool.getConnection();
            cstmt = sqlConn.prepareCall("{call setAudit(?,?,?,?,?,?,?,?,?)}");
            cstmt.setDate(1, stDate);
            cstmt.setDate(2, enDate);
            cstmt.setString(3, siteId);
            cstmt.setString(4, userIen);
            cstmt.setString(5, userNPI);
            cstmt.setString(6, userName);
            cstmt.setString(7, PatId);
            cstmt.setString(8, Category);
            cstmt.setString(9, requestingApp);
            cstmt.executeUpdate();

        } catch (Exception e) {
            LOGGER.error("JMeadows::auditTrack; Exception - " + e.getMessage());
            throw new JMeadowsException(e);

        } finally {
            try {
                if (sqlConn != null) {
                    sqlConn.setAutoCommit(true);
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (cstmt != null) {
                    cstmt.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (sqlConn != null) {
                    sqlConn.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

    /**
     * Sends audit log to encrypted audit sql database.
     *
     * @param siteId Site identifier
     * @param userIen user (user) identifier
     * @param userNPI user (user) national user identifier
     * @param userName user (user) name
     * @param PatId  patient ien.
     * @param Category audit category
     * @param requestingApp requesting application
     * @param StartDate  query start date (defaults to T)
     * @param EndDate    query end date   (defaults to T)
     * @throws JMeadowsException if an error occurs
     */
    public void setAuditEncrypted(String siteId, String userIen, String userNPI,
                         String userName, String PatId, String Category,
                         String requestingApp, String StartDate, String EndDate) throws JMeadowsException {
        Connection sqlConn = null;
        CallableStatement cstmt = null;

        java.sql.Date stDate = null;
        java.sql.Date enDate = null;

        if (Utils.getDate(StartDate) != null) {
            stDate = new java.sql.Date(Utils.getDate(StartDate).getTime());
        }
        if (Utils.getDate(EndDate) != null) {
            enDate = new java.sql.Date(Utils.getDate(EndDate).getTime());
        }

        try {
            sqlConn = sm_dbConnectionPoolEncrypted.getConnection();
            cstmt = sqlConn.prepareCall("{call setAudit(?,?,?,?,?,?,?,?,?)}");
            cstmt.setDate(1, stDate);
            cstmt.setDate(2, enDate);
            cstmt.setString(3, siteId);
            cstmt.setString(4, userIen);
            cstmt.setString(5, userNPI);
            cstmt.setString(6, userName);
            cstmt.setString(7, PatId);
            cstmt.setString(8, Category);
            cstmt.setString(9, requestingApp);
            cstmt.executeUpdate();

        } catch (Exception e) {
            LOGGER.error("JMeadows::auditTrack; Exception - " + e.getMessage());
            throw new JMeadowsException(e);

        } finally {
            try {
                if (sqlConn != null) {
                    sqlConn.setAutoCommit(true);
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (cstmt != null) {
                    cstmt.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (sqlConn != null) {
                    sqlConn.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

    /**
     * Sends audit log to audit sql database.
     *
     * @param siteId Site identifier
     * @param userIen user (user) identifier
     * @param cardID Smart Card ID
     * @param userNPI user (user) national user identifier
     * @param userName user (user) name
     * @param PatId  patient ien.
     * @param Category audit category
     * @param requestingApp requesting application
     * @param StartDate  query start date (defaults to T)
     * @param EndDate    query end date   (defaults to T)
     * @param ipAddress user's IP Address
     * @throws JMeadowsException if an error occurs
     */
    public void setLoginAudit(String cardID, String siteId, String userIen, String userNPI,
                         String userName, String PatId, String Category,
                         String requestingApp, String StartDate, String EndDate, String ipAddress) throws JMeadowsException {
        Connection sqlConn = null;
        CallableStatement cstmt = null;

        java.sql.Date stDate = null;
        java.sql.Date enDate = null;

        if (Utils.getDate(StartDate) != null) {
            stDate = new java.sql.Date(Utils.getDate(StartDate).getTime());
        }
        if (Utils.getDate(EndDate) != null) {
            enDate = new java.sql.Date(Utils.getDate(EndDate).getTime());
        }

        try {
            sqlConn = sm_dbConnectionPool.getConnection();
            cstmt = sqlConn.prepareCall("{call setLoginAudit(?,?,?,?,?,?,?,?,?,?,?)}");
            cstmt.setDate(1, stDate);
            cstmt.setDate(2, enDate);
            cstmt.setString(3, siteId);
            cstmt.setString(4, userIen);
            cstmt.setString(5, cardID);
            cstmt.setString(6, userNPI);
            cstmt.setString(7, userName);
            cstmt.setString(8, PatId);
            cstmt.setString(9, Category);
            cstmt.setString(10, requestingApp);
            cstmt.setString(11, ipAddress);
            cstmt.executeUpdate();

        } catch (Exception e) {
            LOGGER.error("JMeadows::auditTrack; Exception - " + e.getMessage());
            throw new JMeadowsException(e);

        } finally {
            try {
                if (sqlConn != null) {
                    sqlConn.setAutoCommit(true);
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (cstmt != null) {
                    cstmt.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (sqlConn != null) {
                    sqlConn.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

}
