/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsConstants;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartOrder;
import gov.va.med.vds.webservice.FreeTextReport;
import gov.va.med.vds.webservice.Order;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vhahon.common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;

/**
 *
 * Patient orders data access object.
 */
public class OrdersDao extends DaoBase {

    /**
     * Retrieves a list of patient orders from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return A list of patient orders.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Order> getPatientOrders(JMeadowsQuery queryBean) throws JMeadowsException {
        List<Order> resultList = new ArrayList<Order>();

        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatientOrders", queryBean);
        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                MultiLocationQueryConfig multiLocOrdersConfig =
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientOrders")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapOrdersFromBHIE")
                                .location(bhieLoc)
                                .build();

//                MultiLocationQueryConfig multiLocConsultOrdersConfig =
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_bhieRelayServiceHandler)
//                                .dataServiceMethod("getPatientConsultOrders")
//                                .dataServiceMethodArgs(bhieQueryBean)
//                                .smartBeanMappingMethod("mapConsultOrdersFromBHIE")
//                                .location(bhieLoc)
//                                .build();
//

//                MultiLocationQueryConfig multiLocLabOrdersConfig =
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_bhieRelayServiceHandler)
//                                .dataServiceMethod("getPatientLabOrders")
//                                .dataServiceMethodArgs(bhieQueryBean)
//                                .smartBeanMappingMethod("mapLabOrdersFromBHIE")
//                                .location(bhieLoc)
//                                .build();
//
//                MultiLocationQueryConfig multiLocMedicationOrdersConfig =
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_bhieRelayServiceHandler)
//                                .dataServiceMethod("getPatientMedicationOrders")
//                                .dataServiceMethodArgs(bhieQueryBean)
//                                .smartBeanMappingMethod("mapMedicationOrdersFromBHIE")
//                                .location(bhieLoc)
//                                .build();
//
//                MultiLocationQueryConfig multiLocRadiologyOrdersConfig =
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_bhieRelayServiceHandler)
//                                .dataServiceMethod("getPatientRadiologyOrders")
//                                .dataServiceMethodArgs(bhieQueryBean)
//                                .smartBeanMappingMethod("mapRadiologyOrdersFromBHIE")
//                                .location(bhieLoc)
//                                .build();

                configs.add(multiLocOrdersConfig);
//                configs.add(multiLocConsultOrdersConfig);
//                configs.add(multiLocLabOrdersConfig);
//                configs.add(multiLocMedicationOrdersConfig);
//                configs.add(multiLocRadiologyOrdersConfig);
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getPatientOrders")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        Object[] orderObjs = MultiLocationQuery.execute(configs,
                new SmartOrder(), queryBean);

        List<Order> ordersList = null;

        if (orderObjs != null) {
            ordersList = Arrays.asList((Order []) orderObjs);
            // add logic to map various coding systems rxNorm, Loinc
            List<Order> consultList = new ArrayList<Order>();
            List<Order> otherList = new ArrayList<Order>();
            List<Order> labList = new ArrayList<Order>();
            List<Order> medicationList = new ArrayList<Order>();
            List<Order> radiologyList = new ArrayList<Order>();
            for (Order order : ordersList) {
                if (order.getSite() != null && AGENCY_DOD.equalsIgnoreCase(order.getSite().getAgency())) {
                    if (order.getType() == null) {
                        otherList.add(order);
                    } else {
                        if (JMeadowsConstants.CONSULT.equalsIgnoreCase(order.getType())) {
                            consultList.add(order);
                        } else if (JMeadowsConstants.LAB.equalsIgnoreCase(order.getType())) {
                            labList.add(order);
                        } else if (JMeadowsConstants.MEDICATION.equalsIgnoreCase(order.getType())) {
                            medicationList.add(order);
                        } else if (JMeadowsConstants.RADIOLOGY.equalsIgnoreCase(order.getType())) {
                            radiologyList.add(order);
                        } else {
                            otherList.add(order);
                        }
                    }
                } else if (AGENCY_VA.equalsIgnoreCase(order.getSite().getAgency())) {
                    if (order.getType() == null) {
                        otherList.add(order);
                    } else {
                        String lowerCaseTypeString = order.getType().toLowerCase(Locale.getDefault());
                        if (lowerCaseTypeString.startsWith(JMeadowsConstants.CONSULT.toLowerCase(Locale.getDefault()))) {
                            consultList.add(order);
                        } else if (lowerCaseTypeString.startsWith(JMeadowsConstants.LAB.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.CHEMISTRY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.HEMATOLOGY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.MICROBIOLOGY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.ANATOMIC_PATHOLOGY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.SURGICAL_PATHOLOGY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.ELECTRON_MICROSCOPY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.CYTOLOGY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.AUTOPSY.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.BLOOD_BANK.toLowerCase(Locale.getDefault()))) {
                            labList.add(order);
                        } else if (lowerCaseTypeString.startsWith(JMeadowsConstants.IMAGING.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.MRI.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.ANGIO_NEURO.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.CARDIOLOGY_STUDIES.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.NUCLEAR_MEDICINE.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.ULTRASOUND.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.VASCULAR_LAB.toLowerCase(Locale.getDefault()))
                                || lowerCaseTypeString.startsWith(JMeadowsConstants.MAMMOGRAPHY.toLowerCase(Locale.getDefault()))) {
                            radiologyList.add(order);
                        } else if (lowerCaseTypeString.contains(JMeadowsConstants.MEDS.toLowerCase(Locale.getDefault()))) {
                            medicationList.add(order);
                        } else {
                            otherList.add(order);
                        }
                    }
                }
            }

            //ignore consults for now
            //ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, consultList, CodingSystem.CONSULT_ORDERS);
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, labList, CodingSystem.LAB_ORDERS);
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, radiologyList, CodingSystem.RADIOLOGY_ORDERS);
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, medicationList, CodingSystem.MEDICATION_ORDERS);

            resultList.addAll(consultList);
            resultList.addAll(labList);
            resultList.addAll(radiologyList);
            resultList.addAll(medicationList);
            resultList.addAll(otherList);
        }

        return resultList;
    }

    /**
     * Retrieves a patient order from the provided record location.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     * recordSiteCode
     * itemId
     *
     * @return Order details string.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public FreeTextReport getOrderDetail(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.PATIENT,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        auditQuery("OrderDetail", queryBean);
        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        FreeTextReport orderDetail = null;

        Site recordSite = getSite(queryBean.getRecordSiteCode());

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())
                    && LocationHelper.vistaLocationsContainSiteCode(recordSite,
                            queryBean.getPatient().getVistaSites())) {
                orderDetail = sm_vistaHandler.getOrderDetail(VistaBeanFactory.create(queryBean));
                break;
            }
        }

        return orderDetail;
    }
    
}