/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dao.provider;

import gov.va.med.bhie.SnareUser;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.AuthUserInfo;
import gov.va.med.jmeadows.dao.patient.DaoBase;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * User access data access object
 */
public class AuthUserDao extends DaoBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(AuthUserDao.class);

    /**
     * Gets user's JLV access entry in database.
     *
     * @param authUserInfoQuery Query parameters:
     *
     * {@link gov.va.med.jmeadows.dao.beans.AuthUserInfo} Required parameters:
     *
     * cardId
     * organization
     *
     * @return User access entry.
     * @throws JMeadowsException if an error occurs during sql query execution.
     */
    public AuthUserInfo getEntry(AuthUserInfo authUserInfoQuery) throws JMeadowsException
    {
        AuthUserInfo rtc;
        Connection sqlConn = null;
        CallableStatement cstmt = null;
        ResultSet rset = null;
        AuthUserInfo authUserInfoResult = null;

        // Check incoming query parameters and ensure required parameters are set
        if (authUserInfoQuery == null
                || authUserInfoQuery.getCardId() == null
                || authUserInfoQuery.getOrganization() == null
                || authUserInfoQuery.getEmail() == null) {
            LOGGER.error("Required authUserInfoQuery parameters are not valid");
            rtc = null;
        } else {
    
            try {
                // Acquire database connection
                sqlConn = sm_dbConnectionPool.getConnection();
    
                // Prepare statement
                cstmt = sqlConn.prepareCall("{call getAuthUser(?,?,?)}");
                cstmt.setString(1, authUserInfoQuery.getCardId());
                cstmt.setString(2, authUserInfoQuery.getOrganization());
                cstmt.setString(3, authUserInfoQuery.getEmail());
    
                // Execute query
                rset = cstmt.executeQuery();
    
                // Retrieve results
                while (rset.next()) {
                    authUserInfoResult = new AuthUserInfo();
                    authUserInfoResult.setCardId(rset.getString("cardId"));
                    authUserInfoResult.setOrganization(rset.getString("organization"));
                    authUserInfoResult.setFacility(rset.getString("facility"));
                    authUserInfoResult.setName(rset.getString("name"));
                    authUserInfoResult.setEmail(rset.getString("email"));
                    authUserInfoResult.setPhone(rset.getString("phone"));
                }
    
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
                throw new JMeadowsException(e);
            } finally {
                // Close result set
                try {
                    if (rset != null) {
                        rset.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }
    
                // Close callable statement
                try {
                    if (cstmt != null) {
                        cstmt.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }
    
                // Close database connection
                try {
                    if (sqlConn != null) {
                        sqlConn.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }
            }
            rtc = authUserInfoResult;
        }
        return rtc;
    }



    /**
     * Updates user subjectDN (from smartcard cert).
     *
     * @param authUserInfoQuery Query parameters:
     *
     * {@link gov.va.med.jmeadows.dao.beans.AuthUserInfo} Required parameters:
     *
     * cardId
     * organization
     * subjectDN
     *
     * @throws JMeadowsException if an error occurs during sql query execution.
     */
    public boolean updateUserSubjectDN(AuthUserInfo authUserInfoQuery) throws JMeadowsException
    {
        boolean rtc;
        Connection sqlConn = null;
        CallableStatement cstmt = null;

        boolean bRet = false;

        // Check incoming query parameters and ensure required parameters are set
        if (authUserInfoQuery == null
                || authUserInfoQuery.getCardId() == null
                || authUserInfoQuery.getOrganization() == null
                || authUserInfoQuery.getSubjectDN() == null) {
            LOGGER.error("Required authUserInfoQuery parameters are not valid");
            rtc = false;
        } else {
    
            try {
                // Acquire database connection
                sqlConn = sm_dbConnectionPool.getConnection();
                sqlConn.setAutoCommit(false);
    
                // Prepare statement
                cstmt = sqlConn.prepareCall("{call updateSubjectDN(?,?,?)}");
                cstmt.setString(1, authUserInfoQuery.getCardId());
                cstmt.setString(2, authUserInfoQuery.getOrganization());
                cstmt.setString(3, authUserInfoQuery.getSubjectDN());
    
                // Execute query
                cstmt.execute();
                sqlConn.commit();
    
                bRet = true;
    
    //            // Retrieve results
    //            while(rset.next()) {
    //                authUserInfoResult = new AuthUserInfo();
    //                authUserInfoResult.setCardId(rset.getString("cardId"));
    //                authUserInfoResult.setOrganization(rset.getString("organization"));
    //                authUserInfoResult.setFacility(rset.getString("facility"));
    //                authUserInfoResult.setName(rset.getString("name"));
    //                authUserInfoResult.setEmail(rset.getString("email"));
    //                authUserInfoResult.setPhone(rset.getString("phone"));
    //            }
    
            } catch (SQLException e) { // Trap Sql Exceptions
                LOGGER.error("JLV UserSubjectDN Update Exception: " + e.getMessage());
    
                try {
                    if (sqlConn != null) {
                        sqlConn.rollback();
                    }
                } catch (SQLException e2) {
                    LOGGER.error("rollback Exception: " + e2.getMessage());
                }
    
                throw new JMeadowsException(e);
    
            } finally {
                try {
                    if (sqlConn != null) {
                        sqlConn.setAutoCommit(true);
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }
    
                try {
                    if (cstmt != null) {
                        cstmt.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }
    
                try {
                    if (sqlConn != null) {
                        sqlConn.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }
            }
            rtc = bRet;
        }
        return rtc;
    }
    
    public SnareUser getSnareworksUser(String certificate, String user, String credential) throws JMeadowsException {
        return sm_bhieRelayServiceHandler.getSnareworksUser(certificate, user, credential);
    }
}
