/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.provider;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.patient.DaoBase;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.ProviderMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.UserLocation;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.provider.SmartAdmission;
import gov.va.med.vds.webservice.PatientAdmission;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * User Admissions data access object.
 */
public class ProviderAdmissionsDao extends DaoBase {

     /**
     * Gets a list of patients who have been admitted or are being attended
     * by the user passed in the queryBean.
     * @param queryBean requires a user
     * @return a list of patients who were admitted or are being attended
     * @throws gov.va.med.jmeadows.common.JMeadowsException when an error occurs in the retrieval process
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientAdmission> getProviderAdmissions(JMeadowsQuery queryBean) throws JMeadowsException {
        List<PatientAdmission> rtc = null;
        validateQueryBean(queryBean, QueryBeanParams.USER);

        auditQuery("ProviderAdmissions", queryBean);

        List<MultiLocationQueryConfig> configs =
                new ArrayList<MultiLocationQueryConfig>();

        List<UserLocation> locations;

        locations = LocationHelper.getUserLocations(queryBean.getUser());

        for (UserLocation location : locations) {
            String systemType = location.getSystemType();
            String method = "getProviderAdmissions";

            if (SYSTEM_TYPE_CHCS.equalsIgnoreCase(systemType)) {
                rtc = Arrays.asList((PatientAdmission[])
                        new SmartAdmission().generateConnectionUnavailableBean(location, PROVIDER_QUERY_NOT_SUPPORTED_DOD));
                break;
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(systemType)) {
                configs.add(
                    new ProviderMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_vistaHandler)
                        .dataServiceMethod(method)
                        .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                        .location(location)
                        .build()
                );
            }
        }
        if (rtc == null) {
            Object[] admObjs = MultiLocationQuery.execute(configs,
                    new SmartAdmission(), queryBean);
    
            List<PatientAdmission> admissionsList = null;
    
            if (admObjs != null) {
                admissionsList = Arrays.asList((PatientAdmission []) admObjs);
            }
    
            rtc = admissionsList;
        }
        return rtc;
    }

    /**
     * Gets a list of patients who are admitted in the specified ward(s). This
     * method <b>only</b> queries the local VistA system.
     * @param queryBean requires itemId. itemId is a ^ delimited ward ien
     * string, ie "24^47" for two wards or just "24" for a single ward.
     * @return a list of patients in the specified ward(s)
     * @throws gov.va.med.jmeadows.common.JMeadowsException when an error occurs in the retrieval process
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientAdmission> getWardAdmissions(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.ITEM_ID);

        auditQuery("WardAdmissions", queryBean);

        List<MultiLocationQueryConfig> configs =
                new ArrayList<MultiLocationQueryConfig>();

        List<UserLocation> locations;

        locations = LocationHelper.getUserLocations(queryBean.getUser());

        for (UserLocation location : locations) {

            String systemType = location.getSystemType();
            String method = "getWardAdmissions";
            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(systemType)) {
                configs.add(
                    new ProviderMultiLocQueryCfgBuilder()
                        .dataServiceHandler(sm_vistaHandler)
                        .dataServiceMethod(method)
                        .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                        .location(location)
                        .build()
                );
            }
        }

        Object[] admObjs = MultiLocationQuery.execute(
                configs, new SmartAdmission(), queryBean);

        List<PatientAdmission> admissionsList = null;

        if (admObjs != null) {
            admissionsList = Arrays.asList((PatientAdmission []) admObjs);
        }

        return admissionsList;
    }

}
