/*
 * Janus 4.0 (c)
 * Copyright (c) 2012 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dataservicehandler;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.PDWSQueryBean;
import gov.va.med.jmeadows.dao.beans.ResponsePatientQuery;
import gov.va.med.pdws.*;
import gov.va.med.vds.webservice.Patient;
import gov.va.med.vds.webservice.User;

import java.io.Serializable;
import java.math.BigInteger;
import java.net.MalformedURLException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class EntityPatientRetrieveServiceHandler extends DataServiceHandler {
    private static final Logger LOGGER = LoggerFactory.getLogger(EntityPatientRetrieveServiceHandler.class);
    private static final String AGENCY_DOD = APP_CONFIG.getAgencyDOD();
    private static final String AGENCY_VA = APP_CONFIG.getAgencyVA();
    private EntityPatientInquiry entityPatientInquiry;

    public EntityPatientRetrieveServiceHandler(String url) {
        super();
        serviceURL = url;
        serviceName = "PDWSPatientRetrieve";
        serviceNamespace = "http://webservice.pdws.domain.ext/";
        qServiceName = new QName(serviceNamespace, serviceName);

        entityPatientInquiry = new EntityPatientInquiry();
    }

    private EntityPatientRetrievePortType getEntityPatientRetrievePort() throws MalformedURLException {
        EntityPatientRetrievePortType port;
        String appEnv = APP_CONFIG.getAppEnvironment();

        port = entityPatientInquiry.getEntityPatientRetrievePortSoap11();
        BindingProvider prov = (BindingProvider) port;
        prov.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, serviceURL);
        prov.getRequestContext().put("com.sun.xml.internal.ws.connect.timeout", serviceConnectionTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.internal.ws.request.timeout", serviceRequestTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.ws.connect.timeout", serviceConnectionTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.ws.request.timeout", serviceRequestTimeoutMS);

        // SOAP message logging - Enable to show logging in output window
        if ("test".equalsIgnoreCase(appEnv) || "development".equalsIgnoreCase(appEnv)) {
            prov.getBinding().getHandlerChain().add(new SOAPLoggingHandler());
            List<Handler> handlerList = prov.getBinding().getHandlerChain();
            handlerList.add(new SOAPLoggingHandler());
            prov.getBinding().setHandlerChain(handlerList);
        }

        return port;
    }

    public ResponsePatientQuery lookupPatient(PDWSQueryBean queryBean) throws JMeadowsException, MalformedURLException {
        PRPAIN201305UV02 request = GenerateRequest(queryBean);
        PRPAIN201306UV02 response = null;
        ResponsePatientQuery rtc = null;

        response = getEntityPatientRetrievePort().entityPatientRetrieve(request);

        if (response != null) {
            rtc = ParseResponseToPatientList(response);
            if (rtc != null && !rtc.getPatients().isEmpty()) {
                boolean found = true;
                while (found) {
                    found = false;
                    List<Patient> patients = rtc.getPatients();
                    Patient blacklisted = null;
                    for (Patient patient : patients) {
                        if (PDWSBlackList.getInstance().isBlackListed(patient.getEDIPI())) {
                            found = true;
                            blacklisted = patient;
                            break;
                        }
                    }
                    if (blacklisted != null) {
                        LOGGER.error("removing blacklisted patient");
                        patients.remove(blacklisted);
                        rtc.setPatients(patients);
                    }
                }
            }
        }

        return rtc;
    }

    private PRPAIN201305UV02 GenerateRequest(PDWSQueryBean queryBean) {
        MCCIMT000100UV01Device device;
        II ii;
        CS cs;
        CD cd;

        User user = queryBean.getUser();
        String agency = user.getAgency();
        String dmisId = user.getHostSite().getDmisId();
        String siteId = user.getHostSite().getSiteCode();

        ObjectFactory factory = new ObjectFactory();
        PRPAIN201305UV02 request = factory.createPRPAIN201305UV02();

        //id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.2017");
        ii.setExtension("DMDC");
        request.setId(ii);

        //creationTime
        TS ts;
        ts = factory.createTS();
        SimpleDateFormat simpleDateFormat;
        simpleDateFormat = new SimpleDateFormat("yyyyMMddhhmmss");
        ts.setValue(simpleDateFormat.format(new Date()));
        request.setCreationTime(ts);

        //versionCode (optional?)
        cs = factory.createCS();
        cs.setCode("V3PR1");
        request.setVersionCode(cs);

        //interactionId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.1.6");
        ii.setExtension("PRPA_IN201306UV02");
        request.setInteractionId(ii);

        //processingCode
        cs = factory.createCS();
        cs.setCode(APP_CONFIG.getPDWSProcCode());
        request.setProcessingCode(cs);

        //processingModeCode
        cs = factory.createCS();
        cs.setCode("I");
        request.setProcessingModeCode(cs);

        //acceptAckCode
        cs = factory.createCS();
        cs.setCode("AL");
        request.setAcceptAckCode(cs);

        //receiver
        MCCIMT000100UV01Receiver receiver = factory.createMCCIMT000100UV01Receiver();
        request.getReceiver().add(receiver);

        //receiver/typeCode
        receiver.setTypeCode(CommunicationFunctionType.RCV);

        //receiver/device
        device = factory.createMCCIMT000100UV01Device();
        receiver.setDevice(device);

        //receiver/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //receiver/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //receiver/device/id
        ii = factory.createII();
        ii.setRoot("1.3.6.1.4.1.21367.2009.1.3.1.145");
        device.getId().add(ii);

        //sender
        MCCIMT000100UV01Sender sender = factory.createMCCIMT000100UV01Sender();
        request.setSender(sender);

        //sender/typeCode
        sender.setTypeCode(CommunicationFunctionType.SND);

        //sender/device
        device = factory.createMCCIMT000100UV01Device();
        sender.setDevice(device);

        //sender/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //sender/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //sender/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.2017.11.7");
        ii.setExtension(APP_CONFIG.getPDWSSndDevId());
        device.getId().add(ii);

        //sender/device/asAgent
        MCCIMT000100UV01Agent agent = factory.createMCCIMT000100UV01Agent();
        device.setAsAgent(factory.createMCCIMT000100UV01DeviceAsAgent(agent));

        //sender/device/asAgent/classCode
        agent.getClassCode().add("AGNT");

        //sender/device/asAgent/representedOrganization
        MCCIMT000100UV01Organization organization = factory.createMCCIMT000100UV01Organization();
        agent.setRepresentedOrganization(factory.createMCCIMT000100UV01AgentRepresentedOrganization(
                organization));

        //sender/device/asAgent/representedOrganization/classCode
        organization.setClassCode("ORG");

        //sender/device/asAgent/representedOrganization/determinerCode
        organization.setDeterminerCode("INSTANCE");

        //sender/device/asAgent/representedOrganization/id
        ii = factory.createII();
        if (AGENCY_DOD.equalsIgnoreCase(agency)) {
            ii.setRoot("2.16.840.1.113883.3.42.10002.100001.8");
            ii.setExtension(dmisId);
        } else if (AGENCY_VA.equalsIgnoreCase(agency)) {
            ii.setRoot("2.16.840.1.113883.4.349");
            ii.setExtension(siteId);
        }
        organization.getId().add(ii);

        //controlActProcess
        PRPAIN201305UV02QUQIMT021001UV01ControlActProcess controlActProcess = factory.createPRPAIN201305UV02QUQIMT021001UV01ControlActProcess();
        request.setControlActProcess(controlActProcess);

        //controlActProcess/classCode
        controlActProcess.setClassCode(ActClassControlAct.CACT);

        //controlActProcess/moodCode
        controlActProcess.setMoodCode(XActMoodIntentEvent.EVN);

        //controlActProcess/code
        cd = factory.createCD();
        cd.setCode("PRPA_TE201305UV02");
        controlActProcess.setCode(cd);

        //controlActProcess/codeSystem
        cd.setCodeSystem("2.16.840.1.113883.1.6");

        //controlActProcess/dataEnterer
        QUQIMT021001UV01DataEnterer dataEnterer = factory.createQUQIMT021001UV01DataEnterer();
        controlActProcess.getDataEnterer().add(dataEnterer);

        //controlActProcess/dataEnterer/typeCode
        dataEnterer.getTypeCode().add("ENT");

        //controlActProcess/dataEnterer/ contextControlCode
        dataEnterer.setContextControlCode("AP");

        //controlActProcess/dataEnterer/assignedPerson
        COCTMT090100UV01AssignedPerson assignedPerson = factory.createCOCTMT090100UV01AssignedPerson();
        dataEnterer.setAssignedPerson(assignedPerson);

        //controlActProcess/dataEnterer/assignedPerson/classCode
        assignedPerson.setClassCode("ASSIGNED");

        //controlActProcess/dataEnterer/assignedPerson/id
        ii = factory.createII();
        if (AGENCY_DOD.equalsIgnoreCase(agency)) {
            ii.setRoot("2.16.840.1.113883.3.42.127.100001.79");
            ii.setExtension(user.getUserIen() + "^" + dmisId);
        } else if (AGENCY_VA.equalsIgnoreCase(agency)) {
            ii.setRoot("2.16.840.1.113883.4.349");
            ii.setExtension(user.getUserIen() + "^" + siteId);
        }
        assignedPerson.getId().add(ii);

        //controlActProcess/dataEnterer/assignedPerson/Person
        COCTMT090100UV01Person person = factory.createCOCTMT090100UV01Person();
        assignedPerson.setAssignedPerson(factory.createCOCTMT090100UV01AssignedPersonAssignedPerson(person));

        //controlActProcess/dataEnterer/assignedPerson/Person/name
        EN en = factory.createEN();
        en.getContent().add(queryBean.getUser().getName());
        person.getName().add(en);

        //controlActProcess/queryByParameter
        PRPAMT201306UV02QueryByParameter queryByParameter = factory.createPRPAMT201306UV02QueryByParameter();
        controlActProcess.setQueryByParameter(factory.createPRPAIN201306UV02MFMIMT700711UV01ControlActProcessQueryByParameter(queryByParameter));

        //controlActProcess/queryByParameter/queryId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.2017");
        ii.setExtension("333334433");
        queryByParameter.setQueryId(ii);

        //controlActProcess/queryByParameter/statusCode
        cs = factory.createCS();
        cs.setCode("new");
        queryByParameter.setStatusCode(cs);

        //controlActProcess/queryByParameter/initialQuantity
        INT i = factory.createINT();
        i.setValue(BigInteger.valueOf(10));
        queryByParameter.setInitialQuantity(i);

        //controlActProcess/queryByParameter/parameterList
        PRPAMT201306UV02ParameterList parameterList = factory.createPRPAMT201306UV02ParameterList();
        queryByParameter.setParameterList(parameterList);

        //controlActProcess/queryByParameter/parameterList/livingSubjectId
        if (!"".equals(queryBean.getPatientId()) && !"".equals(queryBean.getPatientIdType()))
        {
            ii = factory.createII();
            ii.setRoot("2.16.840.1.113883.3.42.10001.100001.12");
            ii.setExtension(queryBean.getPatientId());

            PRPAMT201306UV02LivingSubjectId livingSubjectId = factory.createPRPAMT201306UV02LivingSubjectId();
            livingSubjectId.getValue().add(ii);
            parameterList.getLivingSubjectId().add(livingSubjectId);

            //controlActProcess/queryByParameter/parameterList/livingSubjectId/semanticsText
            ST st = factory.createST();
            st.getContent().add("LivingSubject.id");
            livingSubjectId.setSemanticsText(st);
        }

        return request;
    }

    private ResponsePatientQuery ParseResponseToPatientList(PRPAIN201306UV02 response) throws JMeadowsException {
        ResponsePatientQuery respPatQuery = new ResponsePatientQuery();
        boolean done = false;

        // Initially set to failure with generic error message. Will only be
        // set to success if there are no acknowledgement errors and no
        // errors retrieving patient list, even if the list is empty
        respPatQuery.setSuccess(false);
        respPatQuery.setErrorMsg("Error retrieving patients."
                + "<br><br>Please try again."
                + "<br><br>If the problem persists, notify your system administrator or help desk.");

        // Create list of result patients
        List<Patient> resultPatList = new ArrayList<Patient>();

        if (response != null) {
            // Check the response acknowledgement for an application error
            List<MCCIMT000300UV01Acknowledgement> ackList = response.getAcknowledgement();
            if (ackList != null) {
                for (MCCIMT000300UV01Acknowledgement ack : ackList) {
                    CS cs = ack.getTypeCode();
                    if (cs != null && "AE".equalsIgnoreCase(cs.getCode())) {
                        // Application error occurred
                        List<MCCIMT000300UV01AcknowledgementDetail> ackDetailList = ack.getAcknowledgementDetail();
                        if (ackDetailList != null) {
                            for (MCCIMT000300UV01AcknowledgementDetail ackDetail : ackDetailList) {
                                if (ackDetail.getTypeCode() == AcknowledgementDetailType.E) {
                                    // Application error message
                                    respPatQuery.setSuccess(false);

                                    CE ce = ackDetail.getCode();
                                    if ((ce != null) && (ce.getDisplayName() != null))
                                    {
                                        LOGGER.error(ce.getDisplayName());
                                        respPatQuery.setErrorMsg(ce.getDisplayName());
                                    }

                                    done = true;
                                    break;
                                }
                            }
                        }
                    }
                }
            }

            if (!done) {
                // Parse message for list of patients
                PRPAIN201306UV02MFMIMT700711UV01ControlActProcess ctrlActProc = response.getControlActProcess();
                if (ctrlActProc != null) {
                    // Iterate through each patient
                    List<PRPAIN201306UV02MFMIMT700711UV01Subject1> subjList = ctrlActProc.getSubject();
                    if (subjList != null) {
                        StringBuilder familyName = new StringBuilder();
                        StringBuilder givenName = new StringBuilder();
                        StringBuilder prefixName = new StringBuilder();
                        StringBuilder suffixName = new StringBuilder();
                        for (PRPAIN201306UV02MFMIMT700711UV01Subject1 subj : subjList) {
                            familyName.setLength(0);
                            givenName.setLength(0);
                            prefixName.setLength(0);
                            suffixName.setLength(0);
                            String edipi = null;
                            String ssn = "N/A";
                            String gender = "N/A";
                            String dob = "N/A";
    
                            PRPAIN201306UV02MFMIMT700711UV01RegistrationEvent regEvent = subj.getRegistrationEvent();
                            if (regEvent != null) {
                                PRPAIN201306UV02MFMIMT700711UV01Subject2 subj1 = regEvent.getSubject1();
                                if (subj1 != null) {
                                    PRPAMT201310UV02Patient pat = subj1.getPatient();
                                    if (pat != null) {
                                        // Retrieve patient EDIPI
                                        List<II> iiList = pat.getId();
                                        if ((iiList != null) && !iiList.isEmpty()) {
                                            edipi = iiList.get(0).getExtension();
                                        }
    
                                        // Retrieve patient's demographics information (name, gender, dob, ssn)
                                        JAXBElement<PRPAMT201310UV02Person> jaxbElementPerson = pat.getPatientPerson();
                                        if (jaxbElementPerson != null) {
                                            PRPAMT201310UV02Person person = jaxbElementPerson.getValue();
                                            if (person != null) {
                                                // Retrieve patient gender
                                                CE ce = person.getAdministrativeGenderCode();
                                                if (ce != null) {
                                                    gender = ce.getCode();
                                                }
    
                                                // Retrieve patient date of birth
                                                TS ts = person.getBirthTime();
                                                if (ts != null) {
                                                    dob = ts.getValue();
                                                }
    
                                                // Retrieve patient name
                                                List<PN> pnList = person.getName();
                                                if (pnList == null) {
                                                    LOGGER.error("List<PN> is null");
                                                } else {
                                                    for (PN pn : pnList) {
                                                        List<Serializable> serialList = pn.getContent();
                                                        if (serialList == null) {
                                                            LOGGER.error("List<Serializable> is null");
                                                        } else {
                                                            for (Serializable serial : serialList) {
                                                                if (serial instanceof JAXBElement) {
                                                                    Object obj = ((JAXBElement) serial).getValue();
                                                                    if (obj != null) {
                                                                        if (obj instanceof EnFamily) {
                                                                            List<Serializable> serial2List = ((EnFamily) obj).getContent();
                                                                            if ((serial2List != null) && !serial2List.isEmpty()) {
                                                                                familyName.append(serial2List.get(0) + " ");
                                                                            }
                                                                        } else if (obj instanceof EnGiven) {
                                                                            List<Serializable> serial2List = ((EnGiven) obj).getContent();
                                                                            if ((serial2List != null) && !serial2List.isEmpty()) {
                                                                                givenName.append(serial2List.get(0) + " ");
                                                                            }
                                                                        } else if (obj instanceof EnPrefix) {
                                                                            List<Serializable> serial2List = ((EnPrefix) obj).getContent();
                                                                            if ((serial2List != null) && !serial2List.isEmpty()) {
                                                                                prefixName.append(serial2List.get(0) + " ");
                                                                            }
                                                                        } else if (obj instanceof EnSuffix) {
                                                                            List<Serializable> serial2List = ((EnSuffix) obj).getContent();
                                                                            if ((serial2List != null) && !serial2List.isEmpty()) {
                                                                                suffixName.append(serial2List.get(0) + " ");
                                                                            }
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
    
                                                    // Format: PrefixName FamilyName,GivenName SuffixName
                                                    // (comma will be added when generating name string)
    
                                                    // Add prefix before family name
                                                    if (prefixName.length() > 0) {
                                                        familyName = prefixName.append(familyName);
                                                    }
    
                                                    // Add suffix after given name
                                                    if (suffixName.length() > 0) {
                                                        givenName.append(suffixName);
                                                    }
                                                }
    
                                                // Retrieve patient SSN
                                                List<PRPAMT201310UV02OtherIDs> otherIdsList = person.getAsOtherIDs();
                                                if (otherIdsList != null) {
                                                    for (PRPAMT201310UV02OtherIDs otherIDs : otherIdsList)
                                                    {
                                                        List<II> ii2List = otherIDs.getId();
                                                        if (ii2List != null) {
                                                            for (II ii : ii2List) {
                                                                CS cs = otherIDs.getStatusCode();
                                                                if (cs != null && "2.16.840.1.113883.4.1".equalsIgnoreCase(ii.getRoot()) && "active".equalsIgnoreCase(cs.getCode())) {
                                                                    ssn = ii.getExtension();
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
    
                            Patient resultPat = new Patient();
    
                            // Set patient information
                            resultPat.setEDIPI(edipi);
                            resultPat.setSSN(ssn);
                            resultPat.setName(familyName.toString().trim() + "," + givenName.toString().trim());
                            resultPat.setGender(gender);
                            resultPat.setDob(dob);
    
                            // Add to patient list
                            resultPatList.add(resultPat);
                        }
                    }
                }
            }
    
            respPatQuery.setSuccess(true);
            respPatQuery.setErrorMsg(null);
            respPatQuery.setPatients(resultPatList);
        }
        return respPatQuery;
    }
}